// Copyright (C) 1996 Fu-Ren Lin
// CapacityPlanning.m

#import "CapacityPlanning.h"

@implementation CapacityPlanning
+create: (id) aZone {
 CapacityPlanning * newCapP;
 newCapP = [super create: aZone];
 newCapP->usedCapList = [List create: aZone];

 return newCapP;
}

-initialCap {
 int i;
 Capacity * aCap;

 for (i=0; i<endDay*2.0; i++) {
   aCap = [Capacity create: [self getZone]];
   aCap->startTime = i*UNIT;
   aCap->endTime = aCap->startTime + UNIT;
   aCap->usedCap   = 0.0;
   aCap->fullMark  = FALSE;
   [self->usedCapList addLast: aCap];
 }
 return self;
}   

//----Capacity planning for coordinate with MRP and ProductionPlanning --
-checkCapacity: (Order *) mfgOrder {
Plan * aPlan;
id <List> planList;
int block, offset;
Capacity * aCap;
int Qleft, assignedQ;
int exhausted = FALSE;
int outOfBoundary = FALSE;
float neededCap;

 [traceAgent pHeader: [scnEntity getTC] value: "capacityP: check" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getScnId] fmt:NL];

 [traceAgent pHeader: [scnEntity getTC] value: "mfgOrder to capacityP=" fmt: Space];
 [traceAgent pOrder:[scnEntity getTC] value: mfgOrder];

 neededCap = [prdPlanningAgent checkUsedCap: mfgOrder->itemId size: mfgOrder->itemQ];
 planList = [List create: [self getZone]];

 if (mfgOrder->dueDate < 0)         // ensure duedate > 0
   mfgOrder->dueDate = mfgOrder->ordArrTime;

 block = [self pointToUsedCapList: mfgOrder->dueDate]; // point at the latest plan unit.
 offset = block;
 
 [traceAgent pHeader: [scnEntity getTC] value: "init-offset=" fmt: Space];
 [traceAgent pInt:[scnEntity getTC] value: offset fmt: NL];

 neededCap = [prdPlanningAgent checkUsedCap: mfgOrder->itemId size: mfgOrder->itemQ];
 if (offset >= [self->usedCapList getCount]) // too far of planning
   return planList;

 aCap = [self->usedCapList atOffset: block];
 Qleft = mfgOrder->itemQ;

 while ((Qleft > 0) && (exhausted == FALSE)) {
   [traceAgent pHeader: [scnEntity getTC] value: "0-offset=" fmt: Space];
   [traceAgent pInt:[scnEntity getTC] value: offset fmt: NL];
 
   if (aCap->fullMark == TRUE) {          // no cap left
     offset--;
     if (offset >= 0)
       aCap = [self->usedCapList atOffset: offset];
     else
       exhausted = TRUE;
   }
   else if (neededCap <= (UNIT - aCap->usedCap)) { // all are in thisblock
     [traceAgent pHeader: [scnEntity getTC] value: "lst-offset=" fmt: Space];
     [traceAgent pInt:[scnEntity getTC] value: offset fmt: NL];

     aPlan = [Plan create: [self getZone]];
     aPlan->prodId = mfgOrder->itemId;
     aPlan->prodQ  = Qleft;
     aPlan->scheduledStart = aCap->startTime;
     aPlan->finishedTime = aCap->startTime + UNIT;
     aPlan->mrpMark = offset;

     [traceAgent pHeader: [scnEntity getTC] value: "mrpMark" fmt: Space];
     [traceAgent pInt: [scnEntity getTC] value: aPlan->mrpMark fmt: NL]; 

 
     [planList addFirst: aPlan];

     [traceAgent pHeader: [scnEntity getTC] value: "aCapPlan-1st" fmt: Space];
     [traceAgent pPlan: [scnEntity getTC] value: aPlan]; 

     Qleft = -1;
   }
   else if (neededCap > (UNIT - aCap->usedCap)) {
     [traceAgent pHeader: [scnEntity getTC] value: "2nd-offset=" fmt: Space];
     [traceAgent pInt:[scnEntity getTC] value: offset fmt: NL];
assignedQ = Qleft * ((UNIT - aCap->usedCap)/neededCap);
     Qleft -= assignedQ;
     aPlan = [Plan create: [self getZone]];
     aPlan->prodId = mfgOrder->itemId;
     aPlan->prodQ  = assignedQ;
     aPlan->scheduledStart = aCap->startTime;
     aPlan->finishedTime = aCap->startTime + UNIT;
     aPlan->mrpMark = offset;
     [planList addFirst: aPlan];  

     [traceAgent pHeader: [scnEntity getTC] value: "aCapPlan-2nd" fmt: Space];
     [traceAgent pPlan: [scnEntity getTC] value: aPlan]; 

     offset--;
     neededCap -= (UNIT - aCap->usedCap); 

     if (offset >= 0)
       aCap = [self->usedCapList atOffset: offset];
     else
       exhausted = TRUE;
   }
 }

  [traceAgent pHeader: [scnEntity getTC] value: "needToLookDown" fmt: NL];
 
 //need to allocate block after dueDate
 if (Qleft <= 0)   
   return planList;
 else  {
   offset = block + 1;
   
   if (offset >= [self->usedCapList getCount]) { // too far of planning
     outOfBoundary = TRUE;
     return planList;
   }
   
   aCap = [self->usedCapList atOffset: offset];
   while ((Qleft > 0) && (outOfBoundary == FALSE)) {
     if (aCap->fullMark == TRUE) { // no cap left
       offset++;
     
       if (offset >= [self->usedCapList getCount]) // too far of planning
         outOfBoundary = TRUE;
       else
         aCap = [self->usedCapList atOffset: offset];
     }
     else if (neededCap <= (UNIT - aCap->usedCap)) { // all are in this block
       aPlan = [Plan create: [self getZone]];
       aPlan->prodId = mfgOrder->itemId;
       aPlan->prodQ  = Qleft;
       aPlan->scheduledStart = aCap->startTime;
       aPlan->finishedTime = aCap->startTime + UNIT;
       aPlan->mrpMark = offset;
       [planList addLast: aPlan];
    
       [traceAgent pHeader: [scnEntity getTC] value: "aCapPlan" fmt: Space];
       [traceAgent pPlan: [scnEntity getTC] value: aPlan]; 
 
       Qleft = -1;
     }
     else if (neededCap > (UNIT - aCap->usedCap)) {
       assignedQ = Qleft * ((UNIT - aCap->usedCap)/neededCap);
       Qleft -= assignedQ;
       aPlan = [Plan create: [self getZone]];
       aPlan->prodId = mfgOrder->itemId;
       aPlan->prodQ  = assignedQ;
       aPlan->scheduledStart = aCap->startTime;
       aPlan->finishedTime = aCap->startTime + UNIT;
       aPlan->mrpMark = offset;
       [planList addLast: aPlan];
  
       [traceAgent pHeader: [scnEntity getTC] value: "aCapPlan" fmt: Space];
       [traceAgent pPlan: [scnEntity getTC] value: aPlan]; 
       neededCap -= (UNIT - aCap->usedCap); 
       offset++;
             
       if (offset >= [self->usedCapList getCount]) // too far of planning
         outOfBoundary = TRUE;
       else 
         aCap = [self->usedCapList atOffset: offset];
     }
   }
 }
    
 return planList;
}

//---allocate capacity for plan
-allocateCapacity: (id) planList {
 id index;
 Plan * capPlan;
 int offset;
 Capacity * aCap;
 float addedLoad;
 
 index = [planList begin: [self getZone]];
 while ((capPlan = [index next])) {
   offset = capPlan->mrpMark;
   aCap = [self->usedCapList atOffset: offset];
   addedLoad = [prdPlanningAgent checkUsedCap: capPlan->prodId size:
capPlan->prodQ];
   aCap->usedCap += addedLoad;
   if ((UNIT - aCap->usedCap) <= (UNIT * 0.05))
    aCap->fullMark = TRUE;
 }
    
 return self;
}

-reviseCapAllocation: (id) oldPlanList earliest: (float) startTime wrt: (int) YesNo {
id <List> newPlanList;
id index;
Plan * oldPlan, * newPlan;
int complete =FALSE;
int revisedQ = 0;
int Qleft, assignedQ;
float neededCap;
int offset;
Capacity * aCap;
int prodNo;
int outOfBoundary = FALSE;

  // accumulate Q needed to reallocate;
 index = [oldPlanList begin: [self getZone]];
 while ((oldPlan = [index next]) && (complete == FALSE)) {
   prodNo = oldPlan->prodId;
   if (oldPlan->scheduledStart < startTime) {
     revisedQ += oldPlan->prodQ;
     [index remove];
   }
   else
     complete = TRUE;
 }
 [index drop];
 
 // rewrite into newPlan
 if ([oldPlanList getCount] > 0) {
   newPlanList = [oldPlanList copy: [self getZone]];
 
   index = [newPlanList begin: [self getZone]];
   while ((newPlan = [index next])) {
     [traceAgent pHeader: [scnEntity getTC] value: "newCapPlan" fmt: Space];
     [traceAgent pPlan: [scnEntity getTC] value: newPlan]; 
   }  
   [index drop];

   newPlan = [newPlanList getLast];
   offset = newPlan->mrpMark + 1;
 }
 else {
   newPlanList = [List create: [self getZone]];
   offset = [self pointToUsedCapList: startTime];
 }
 
 Qleft = revisedQ;
 neededCap = [prdPlanningAgent checkUsedCap: prodNo size: Qleft];

 if (offset >= [self->usedCapList getCount]) // too far of planning
   return newPlanList;

 aCap = [self->usedCapList atOffset: offset];
 while ((Qleft > 0) && (outOfBoundary == FALSE)) {
   if (aCap->fullMark == TRUE) { // no cap left
     offset++;

     if (offset >= [self->usedCapList getCount]) // too far of planning
       outOfBoundary = TRUE;
     else 
       aCap = [self->usedCapList atOffset: offset];
   }
   else if (neededCap <= (UNIT - aCap->usedCap)) { // all are in this block
     newPlan = [Plan create: [self getZone]];
     newPlan->prodId = prodNo;
     newPlan->prodQ  = Qleft;
     newPlan->scheduledStart = aCap->startTime;
     newPlan->finishedTime = aCap->startTime + UNIT;
     newPlan->mrpMark = offset;
     [newPlanList addLast: newPlan];
     Qleft = -1;
   }
   else if (neededCap > (UNIT - aCap->usedCap)) {
     assignedQ = Qleft * ((UNIT - aCap->usedCap)/neededCap);
     Qleft -= assignedQ;
     newPlan = [Plan create: [self getZone]];
     newPlan->prodId = prodNo;
     newPlan->prodQ  = assignedQ;
     newPlan->scheduledStart = aCap->startTime;
     newPlan->finishedTime = aCap->startTime + UNIT;
     newPlan->mrpMark = offset;
     [newPlanList addLast: newPlan];
     neededCap -= (UNIT - aCap->usedCap); 
     offset++;

     if (offset >= [self->usedCapList getCount]) // too far of planning
       outOfBoundary = TRUE;
     else 
       aCap = [self->usedCapList atOffset: offset];
   }
 }
 if (([newPlanList getCount] > 0) && (YesNo == Yes))
   [self allocateCapacity: newPlanList];
   
 return newPlanList;
}

//---obtain the slot at usedCapList --
-(int)pointToUsedCapList: (float) aTime {
 int block;
 id index;
 Capacity * aCap;
 int offset = -1;
 int found = FALSE;

 index = [self->usedCapList begin: [self getZone]];
 while ((aCap = [index next]) && (found == FALSE)) {
   offset++;
   if ((aTime >= aCap->startTime) && (aTime < aCap->endTime))
     found = TRUE;
 }
 [index drop];

 if (found == TRUE)
   block = offset;
 else
   block = Infinite;

 return block;
}

//----------set the agent which processes orders------
-setPrdPlanningAgent: (id) agent {
 prdPlanningAgent = agent;
 return self;
}

-setScnEntity: (id) agent {
 scnEntity = agent;
 return self;
}

-setTraceAgent: (id) agent {
 traceAgent = agent;
 return self;
}

@end



