// Copyright (C) 1996 Fu-Ren Lin
// Dispatching.m

#import "Dispatching.h"

@implementation Dispatching
+create: (id) aZone {
 Dispatching * newDispatch;
 newDispatch = [super create: aZone];
 newDispatch->workOrders = [List create: aZone];
 return newDispatch;
}

-process {
 Plan * workOrder;
 Plan * outWorkOrder;
 int available;
 id index;
 int complete = FALSE;

 [traceAgent pHeader: [scnEntity getTC] value: "dispatch=" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getScnId] fmt:NL];
 // get the workOrder from buildPlan
 while ((workOrder = [prdPlanningAgent getTopWorkOrders]) &&
                                       (complete == FALSE)) {
   if ((workOrder->mrpMark == DONE) && 
       (workOrder->scheduledStart <= getCurrentTime())) { // mrp is done
     [self->workOrders addLast: workOrder];

     [traceAgent pHeader: [scnEntity getTC] value: "workOrder =" fmt: Space];
     [traceAgent pPlan: [scnEntity getTC] value: workOrder];

     [prdPlanningAgent removeTopWorkOrders];
   }
   else
     complete = TRUE;
 }
 
 // dispatch available components from inventory to dispatchList and
 // move workorder to factory
 
 if ([self->workOrders getCount] > 0) {
   index = [self->workOrders begin: [self getZone]];
   while ((workOrder = [index next])) {
     available = [invAgent checkAvailComps: workOrder];
     if (available > 0) {
       outWorkOrder = [workOrder copy: [self getZone]];
       if (available > workOrder->prodQ) {
         outWorkOrder->prodQ  = workOrder->prodQ;
         [index remove];
       }
       else {
         outWorkOrder->prodQ  = available;
         workOrder->prodQ -= available;
       }
       [factoryAgent addDispatchList: outWorkOrder];
       [invAgent     removeToDispatch: outWorkOrder];
       [factoryAgent addDispatchedComps: outWorkOrder];
       [mrpAgent     removeProcessedPlan: outWorkOrder];

       [traceAgent pHeader: [scnEntity getTC] value: "dispatch to shopfloor" fmt: Space];
       [traceAgent pPlan: [scnEntity getTC] value: outWorkOrder];
     }
   }
   [index drop];
 }
 return self;
}

//----
-setPrdPlanningAgent: (id) agent{
 prdPlanningAgent = agent;
 return self;
}

-setMrpAgent: (id) agent {
 mrpAgent = agent;
 return self;
}

-setInvAgent: (id) agent { 
 invAgent = agent;
 return self;
}

-setScnEntity: (id) agent { 
 scnEntity = agent;
 return self;
}

-setFactoryAgent: (id) agent {
 factoryAgent = agent;
 return self;
}


-setTraceAgent: (id) agent {
 traceAgent = agent;
 return self;
}

@end



