// Copyright (C) 1996 Fu-Ren Lin
// Factory.m

#import "Factory.h"

@implementation Factory
+create: (id) aZone {
 Factory * newFactory;
 newFactory = [super create: aZone];
 newFactory->dispatch = [List create: aZone];
 newFactory->dispatchedComps = [List create: aZone];
 newFactory->pWkOrder = [Plan create: aZone];

 return newFactory;
}

//---set the initial state ----
-setInitState {
 self->state = IDLE;
 self->accumuBusyTime = 0.0;
 return self;
}

//----manufacturing parts from dispatchList------
-process {

 [traceAgent pHeader: [scnEntity getTC] value: "factory=" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getScnId] fmt:NL];
 if ((state == BUSY) && (getCurrentTime() >= self->mfgProcEnd)) {
   self->pWkOrder->finishedTime = self->mfgProcEnd;
   self->accumuBusyTime += (mfgProcEnd - self->pWkOrder->scheduledStart);
   [invAgent addInStock: self->pWkOrder];
   
   [traceAgent pHeader: [scnEntity getTC] value: "FINISHED" fmt: Space];
   [traceAgent pPlan: [scnEntity getTC] value: self->pWkOrder];

   [self removeDispatchedComps: self->pWkOrder];

   [prdPlanningAgent fulfillBuildPlan: self->pWkOrder];
   state = IDLE; 
 }
 else if ((state == IDLE) && ([self->dispatch getCount] > 0)) {
   self->pWkOrder = [self removeTopOfDispatch];
   self->pWkOrder->scheduledStart = getCurrentTime(); // for capUtil
   self->mfgProcEnd = getCurrentTime() + [scnEntity getMfgProcTime: self->pWkOrder->prodId size: self->pWkOrder->prodQ];
   state = BUSY;
 }
 return self;
}

-(float) getCapUtilization {
 float capUtil, totalBusyT;
 
 totalBusyT= [self getAccumuBusyTime];
 capUtil = totalBusyT / getCurrentTime();
 
 return capUtil;
}

-(float) getAccumuBusyTime {
 return self->accumuBusyTime;
}

//-----remove top of dispatch list for factory
-(Plan *) removeTopOfDispatch {
Plan * aPlan;
Plan * thePlan;
id index;

  index = [self->dispatch begin:[self getZone]];
  if ((aPlan = [index next])) {
    thePlan = [aPlan copy: [self getZone]];
    [traceAgent pHeader: [scnEntity getTC] value: "remove top plan" fmt: Space];
    [traceAgent pPlan: [scnEntity getTC] value: thePlan];

    [index remove];
    [index drop];
    return thePlan;
  }
  else {
    [index drop];
    return aPlan;
  } 
}    


//-----remove the components for the product ---
-removeDispatchedComps: (Plan *) workOrder {
 id <List> compList;
 id index;
 IdNo * aComp;
 Part * compStock;

 compList = [ofpModel getCompList: workOrder->prodId]; 
 index = [compList begin: [self getZone]];
 while ((aComp = [index next])) {
   compStock = [self allocateDispatchedComps: aComp->no];
   compStock->partQ -= workOrder->prodQ;
 }
 [index drop];
 return self;
}

//-----add the components for the product into dispatchedComps---
-addDispatchedComps: (Plan *) workOrder {
 id <List> compList;
 id index;
 IdNo * aComp;
 Part * compStock;

 compList = [ofpModel getCompList: workOrder->prodId]; 
 index = [compList begin: [self getZone]];
 while ((aComp = [index next])) {
   if ((compStock = [self allocateDispatchedComps: aComp->no])) 
     compStock->partQ += workOrder->prodQ;
   else 
     [self addNewComps: workOrder];
 }
 [index drop];

 return self;
}

-addNewComps: (Plan *) workOrder {
 Part * aPart;
 
 aPart = [Part create: [self getZone]];
 aPart->partId = workOrder->prodId;
 aPart->partQ  = workOrder->prodQ;
 [self->dispatchedComps addLast: aPart];

 return self;
}
//------ point to a part in dispatchedComps
-(Part *) allocateDispatchedComps: (int) identification {
id index;
Part * disComp;
int found = FALSE;
int offset = -1;

 index = [self->dispatchedComps begin: [self getZone]];
 while ((disComp = [index next]) && (found = FALSE)) {
   offset++;
   if (disComp->partId == identification)
     found = TRUE;
 }
 [index drop];
 if (found == TRUE)
   disComp = [self->dispatchedComps atOffset: offset];

 return disComp;
}


//-----------add components to dispatchList at Factory ---
-addDispatchList: (Plan *) workOrder {
 Plan * theWorkOrder;

  theWorkOrder = [workOrder copy: [self getZone]];
  [traceAgent pHeader: [scnEntity getTC] value: "add a wkorder" fmt: Space];
  [traceAgent pPlan: [scnEntity getTC] value: theWorkOrder];

  [self->dispatch addLast: theWorkOrder];
  return self;
}

//----------set the agent which processes orders------
-setInvAgent: (id) agent {
 invAgent = agent;
 return self;
}

//---------------
-setPrdPlanningAgent: (id) agent {
 prdPlanningAgent = agent;
 return self;
}

-setScnEntity: (id) agent {
 scnEntity = agent;
 return self;
}

-setOfpModel: (id) agent {
 ofpModel = agent;
 return self;
}


-setTraceAgent: (id) agent {
 traceAgent = agent;
 return self;
}

@end



