// Copyright (C) 1996 Fu-Ren Lin
// Forecasting.m

#import "Forecasting.h"

@implementation Forecasting

+create: (id) aZone {
 Forecasting * newFore;
 newFore = [super create: aZone];
 newFore->foreDemandL = [List create: aZone];
 return newFore;
}

//----initialize the counterForSchedule --
-initCounterForForecasting {
 self->counterForForecasting = 1;
 return self;
}

//-----------forecasting product demands from downstream entities
-process {
 
// for part souce entity: exit
  if ([self checkTimeToForecasting] == TRUE) { 
    if ([scnEntity getMfgCap] == 'S')
    return self;

   // for BTF only, except for MTO while downstream infoshare is enabled
   if ([scnEntity getDemandPolicy] == BTF)
     [self processBTFMTO: [scnEntity getDownPropLayer]];
   else if (([scnEntity getDemandPolicy] == MTO) &&
               ([scnEntity getCoordID] == Yes))
     [self processBTFMTO: [scnEntity getDownPropLayer]];
  }

 return self;
}

// do forecasting ...
-processBTFMTO: (int) downProp {
 float span;
 float preStart, cTime;
 id <List> preActualL;
 id <List> prodList;
 id <List> preEstDemandL;
 Order * anOrder;
 id index;
 
 [traceAgent pHeader: [scnEntity getTC] value: "forecast=" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getScnId] fmt:NL];

 span = ForecastingH * UNIT; 
 cTime = getCurrentTime();
 preStart = cTime - span;
 if (preStart < 0.0)
   preStart = 0.0;
  
 //obtain the previous actual demand
  prodList = [scnEntity getInvTypes];
  preActualL = [self computeActualDemand: downProp from: preStart to: cTime prodSet: prodList];
   
 //obtain the previous forecasting demand;
   if (self->counterForForecasting == 2) { // increase on in checkTimeToForecasting
     self->foreDemandL = [preActualL copy: [self getZone]];
   }
   else {
     preEstDemandL =  [self->foreDemandL copy: [self getZone]];
     [self cleanForeDemandL];
     self->foreDemandL = [self exponentialF: prodList act: preActualL pred: preEstDemandL];
   }

  //obtain the forecasting demand for the next period
  if ([scnEntity getDemandPolicy] == BTF) {
    index = [self->foreDemandL begin: [self getZone]];
    while ((anOrder = [index next])) {
      anOrder->ordArrTime = cTime;
      anOrder->senderId = [scnEntity getScnId];
      if ([scnEntity getMfgCap] == 'Y') {
        anOrder->wantDate = [prdPlanningAgent estimateFinishTime: anOrder level:[scnEntity getUpPropLayer]];
        anOrder->dueDate  = -1;
        [invAgent addToCompDemandList: anOrder];
        [prdPlanningAgent addToMfgOrders: anOrder];
      }
      else if ([scnEntity getMfgCap] == 'N') {
        anOrder->wantDate = [ordAgent estimateProdArrivalTime: anOrder level: [scnEntity getUpPropLayer]];
        anOrder->dueDate = anOrder->wantDate;
        [invAgent addToCompDemandList: anOrder];
        [ordAgent addOneToOutOrders: anOrder];
      }
      [traceAgent pHeader: [scnEntity getTC] value:"forecasting" fmt: Space];
      [traceAgent pOrder: [scnEntity getTC] value: anOrder];
   }
   [index drop];
 }

 return self; 
}

//---compute the actual demands from four sources list --
-computeActualDemand: (int) downProp from: (float) stime to: (float)etime prodSet: (id) prodList {
 id <List> preActDemandL;
 id <List> preActSentL;
 id <List> preActCompDemandL;
 id <List> preActCompSentL;
 id <List> preActualL;
 id <List> customerList;
 id <List> customerProdList;
 id <List> customerPreActualL;
 id index;
 IdNo * aCustomerId;
 SCNEntitySwarm * aCustomerE;
 int newDownProp;
 float addedTime;

 if (downProp == 0) {
   if ([scnEntity getMfgCap] == 'N') {
     preActDemandL = [invAgent getPreDemand: prodList from: stime to: etime source: [invAgent getDemandList]];
     preActSentL =  [invAgent getPreDemand: prodList from: stime to:
etime source: [invAgent getFulfilledOrders]];

     preActualL  = [List create: [self getZone]];
     preActualL  = [self mergeLists: preActualL with: preActDemandL];
     preActualL  = [self mergeLists: preActualL with: preActSentL];
   }
   else if ([scnEntity getMfgCap] == 'Y') {
     preActDemandL = [invAgent getPreDemand: prodList from: stime to: etime source: [invAgent getDemandList]];
     preActSentL   = [invAgent getPreDemand: prodList from: stime to: etime source: [invAgent getFulfilledOrders]];
     preActCompDemandL = [invAgent getPreDemand: prodList from: stime to: etime source: [invAgent getCompDemandList]];
     preActCompSentL   = [invAgent getPreDemand: prodList from: stime to: etime source: [invAgent getCompFulfilledOrders]];
     
     preActualL  = [List create: [self getZone]];
     preActualL  = [self mergeLists: preActualL with: preActDemandL];
     preActualL  = [self mergeLists: preActualL with: preActSentL];
     preActualL  = [self mergeLists: preActualL with: preActCompDemandL];
     preActualL  = [self mergeLists: preActualL with: preActCompSentL];
   }
 }
 else if (downProp > 0) {
    preActualL = [List create: [self getZone]];
    customerList = [scnEntity getCustomerList];
    index = [customerList begin: [self getZone]];
    while ((aCustomerId = [index next])) {
      aCustomerE = [ofpModel allocateEntity: aCustomerId->no];
      newDownProp = downProp - 1;
      addedTime = [scnEntity getDeliverTime: aCustomerId->no];
      stime += addedTime;
      etime += addedTime;
      customerProdList = [aCustomerE getInvTypes];
      customerPreActualL = [aCustomerE acceptComputeActualDemandQuery:
newDownProp from: stime to: etime prodSet: customerProdList];
      preActualL = [self screenProdSet: customerPreActualL to: preActualL with: prodList];
    }   
    [index drop];
  }
  return preActualL;
}

//-- foreSeeDemand within MTO and BTF entities for query from MTS up entity.
-(int) foreSeeDemand: (DemandQuery *) aQuery {
 int amt = 0;
 float span, period;
 Order * anOrder;
 id index;
 int found = FALSE;

  period = aQuery->endTime - aQuery->startTime;
  span =  ForecastingH * UNIT;  
  
 //obtain the forecasting demand for the next period
   index = [self->foreDemandL begin: [self getZone]];
   while ((anOrder = [index next]) && (found == FALSE)) {
      if (anOrder->itemId == aQuery->prodId) {
        amt = anOrder->itemQ * (period/span);
        found = TRUE;
      }
   }
   [index drop];

 return amt; 
}

//---- exponential forecasting
-exponentialF: (id) prodGroup act: (id) actualL pred: (id) estimatedL {
id index;
id <List> forecastingDL;
Order *fOrder;
IdNo * aProd;
Order * actOrder;
float actNum, foreNum; 
 forecastingDL = [List create: [self getZone]];
 if ([scnEntity getMfgCap] == 'Y') { // according to product group
   index = [prodGroup begin: [self getZone]];
   while ((aProd = [index next])) {
     fOrder = [Order create: [self getZone]];
     fOrder->itemId = aProd->no;
     actNum = [self getQatList: actualL prod: aProd->no];
     foreNum = [self getQatList: estimatedL prod: aProd->no];
     fOrder->itemQ = (Alpha * actNum) + ((1-Alpha)* foreNum);
     // fprintf(fdebug, "fOrder-act: %d, %f\n", fOrder->itemId, actNum);
     // fprintf(fdebug, "fOrder-fore: %d, %f\n", fOrder->itemId, foreNum);
     if (fOrder->itemQ > 0)
       [forecastingDL addLast: fOrder];
   }
   [index drop];
 }
 else if ([scnEntity getMfgCap] == 'N') { // according to actual demand ids
   index = [actualL begin: [self getZone]];
   while ((actOrder = [index next])) {
     fOrder = [Order create: [self getZone]];
     fOrder->itemId = actOrder->itemId;
     actNum  = actOrder->itemQ;
     foreNum = [self getQatList: estimatedL prod: actOrder->itemId];
     fOrder->itemQ = (Alpha * actNum)+((1-Alpha)* foreNum);
     // fprintf(fdebug, "fOrder-act: %d, %f\n", fOrder->itemId, actNum);
     // fprintf(fdebug, "fOrder-fore: %d, %f\n", fOrder->itemId, foreNum);
     if (fOrder->itemQ > 0)
       [forecastingDL addLast: fOrder];
   }
   [index drop];
 }

 return forecastingDL; 
}

//--- get the Quantity of aList with prodId no
-(int)getQatList: (id) aList prod: (int) no {
 id index;
 Order * anOrder;
 int found = FALSE;
 int prodQ = 0;

 index = [aList begin: [self getZone]];
 while ((anOrder = [index next]) && (found == FALSE)) {
   if (anOrder->itemId == no) {
     prodQ = anOrder->itemQ;
     found = TRUE;
   }
 } 
 [index drop];

 return prodQ;
}
    
//----clean the foreDemandL in order to store new forecasting list
-cleanForeDemandL {
id index;
Order * anOrder;
 
 index = [self->foreDemandL begin: [self getZone]];
 while ((anOrder = [index next]))
   [index remove];

 return self;
}

//-----------check the time point for forecasting --
-(int) checkTimeToForecasting {
 int ready;
 int span;

 span = ForecastingH * UNIT;
 if (((counterForForecasting * span) >=  getCurrentTime()) && 
     ((counterForForecasting * span - getCurrentTime()) <= TimeStep)) {
    counterForForecasting++;
   ready = TRUE;
 }
 else
   ready = FALSE;

 return ready;
}

//-- pick the right prod to count on --
-screenProdSet:(id) inList to: (id) outList with: (id) prodList {
 id index1, index2;
 Order * inOrder, * mOrder;
 int found;
 id <List> sList;

 sList = [outList copy: [self getZone]];

 index1 = [inList begin: [self getZone]];
 while ((inOrder = [index1 next]) && ([self isInProdList: inOrder->itemId within: prodList] == Yes)) {
   index2 = [sList begin: [self getZone]];
   found = FALSE;
   while ((mOrder = [index2 next]) && (found == FALSE)) {
     if (inOrder->itemId == mOrder->itemId) {
       mOrder->itemQ += inOrder->itemQ;
       found = TRUE;
     }
    }
    [index2 drop];
    if (found == FALSE) 
      [sList addLast: inOrder];
 }
 [index1 drop];

 return sList;

}

//---check to see the prodNo belong to a prodList --
-(int) isInProdList: (int) prodNo within: (id) prodList {
int belonging;
id index;
IdNo * aProd;
  
  index = [prodList begin: [self getZone]];
  belonging = FALSE;
  while ((aProd = [index next]) && (belonging == FALSE)) {
     if (aProd->no == prodNo)
       belonging = TRUE;
  }

 return belonging;
}

//----sum of the demandList and fulfilledList
-mergeLists: (id) outList with: (id) inList {
 id index1, index2;
 Order * inOrder, * mOrder;
 int found;
 id <List> mList;

 mList = [outList copy: [self getZone]];
 [traceAgent pHeader: [scnEntity getTC] value: "a new inList=" fmt:Tab];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getScnId] fmt: NL];

 index1 = [inList begin: [self getZone]];
 while ((inOrder = [index1 next])) {
 
   [traceAgent pHeader: [scnEntity getTC] value: "inList=" fmt: Space];
   [traceAgent pOrder: [scnEntity getTC] value: inOrder];

   index2 = [mList begin: [self getZone]];
   found = FALSE;
   while ((mOrder = [index2 next]) && (found == FALSE)) {
     if (inOrder->itemId == mOrder->itemId) {
       mOrder->itemQ += inOrder->itemQ;
       found = TRUE;
     }
   }
   [index2 drop];
   if (found == FALSE) 
     [mList addLast: inOrder];
 }
 [index1 drop];

 return mList;
}

-setInvAgent: (id) agent {
 invAgent = agent;
 return self;
}

-setOrdAgent: (id) agent {
 ordAgent = agent;
 return self;
}

-setPrdPlanningAgent: (id) agent {
 prdPlanningAgent = agent;
 return self;
}

-setOfpModel: (id) agent {
 ofpModel = agent;
 return self;
}

-setScnEntity: (id) agent {
 scnEntity = agent;
 return self;
}


-setTraceAgent: (id) agent {
 traceAgent = agent;
 return self;
}

@end
