// Copyright (C) 1996 Fu-Ren Lin
// InventoryManagement.m  is a physical agent handling parts inventory.

#import "InventoryManagement.h"

@implementation InventoryManagement

+create: (id) aZone {
 InventoryManagement *newInvM;
 newInvM = [super create: aZone];
 newInvM->outParts = [List create: aZone];
 newInvM->sentParts = [List create: aZone];
 newInvM->demandList = [List create: aZone];
 newInvM->compDemandList = [List create: aZone];
 newInvM->fulfilledOrders = [List create: aZone];
 newInvM->compFulfilledOrders = [List create: aZone];
 newInvM->receivers = [List create: aZone];
 newInvM->inStock = [List create: aZone];
 newInvM->inPartBuf = [List create: aZone];
 newInvM->ordCycleTs = [List create: aZone];
 newInvM->accuInInvCost = [List create: aZone];
 newInvM->accuOutInvCost = [List create: aZone];
 newInvM->tardyDL= [List create: aZone];
 newInvM->earlyDL = [List create: aZone];
 newInvM->tardyWL= [List create: aZone];
 newInvM->earlyWL = [List create: aZone];

 return newInvM;
}

//---------update the inventory ---
-process {
 id index;
 Part * currentPart, * stockPart, * outPart;
 Order * currentOrder;
 IdNo * receiver;
 Order * fulfilledOrder;

 [traceAgent pHeader: [scnEntity getTC] value: "invM=" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getScnId] fmt:NL];
 // add coming parts into inStock
 while ((currentPart = [self getTopOfPartList: self->inPartBuf])) {
   if ((stockPart = [self allocateStock: currentPart->partId])) 
      stockPart->partQ += currentPart->partQ;
   else 
     [self->inStock addLast: currentPart];
   // update ordM's sentOrders List
   [ordAgent reduceSentOrders: currentPart];
   // reduce the compDemandList
   [self reduceCompDemandList: currentPart->partId size: currentPart->partQ];
   //stats
   [self accumuInputInvCost: currentPart];
   // remove currentPart from inPartBuf
   [self->inPartBuf removeFirst];
 }
 
 // check any orders need to be fulfilled
 index = [self->demandList begin: [self getZone]];
 while ((currentOrder = [index next])) {
  if (currentOrder->wantDate <= getCurrentTime()) {
   if ((currentPart = [self allocateStock: currentOrder->itemId])) {
     if (currentPart->partQ != 0) {
       receiver = [IdNo create: [self getZone]];
       receiver->no = currentOrder->senderId;
       
       if ((currentOrder->itemQ <= currentPart->partQ) &&
                                  (currentPart->partQ != 0)) {
         currentPart->partQ -= currentOrder->itemQ;
         outPart = [Part create: [self getZone]];
         outPart->partId = currentPart->partId;
         outPart->partQ  = currentOrder->itemQ;  //<----- *
         outPart->partArrTime = getCurrentTime();
         [self->outParts addLast: outPart];
         [self->receivers addLast: receiver];
         [self computeOrdCycleTime: currentOrder]; //stats
         [self computeDueDatePerformance: currentOrder];  //stats
         [self->fulfilledOrders addLast: currentOrder];
  
         [traceAgent pHeader: [scnEntity getTC] value:"fulfilledOrder" fmt: Space];
         [traceAgent pOrder: [scnEntity getTC] value: currentOrder];

         [index remove];
       }
       else if ((currentOrder->itemQ > currentPart->partQ) &&
                                  (currentPart->partQ != 0)) {
         currentOrder->itemQ -= currentPart->partQ;
         outPart = [Part create: [self getZone]];
         outPart->partId = currentPart->partId;
         outPart->partQ  = currentPart->partQ;  //<---------*
         outPart->partArrTime = getCurrentTime();
         [self->outParts addLast: outPart];
         [self->receivers addLast: receiver];
         [self computeOrdCycleTime: currentOrder];  //stats
         [self computeDueDatePerformance: currentOrder];  //stats
         
         fulfilledOrder = [Order create: [self getZone]];
         fulfilledOrder->itemId = currentOrder->itemId;
         fulfilledOrder->itemQ = currentPart->partQ;  //<----------------*
         fulfilledOrder->ordArrTime = currentOrder->ordArrTime;
         fulfilledOrder->wantDate = currentOrder->wantDate;
         fulfilledOrder->dueDate = currentOrder->dueDate;
         fulfilledOrder->senderId = currentOrder->senderId;

         [self->fulfilledOrders addLast: fulfilledOrder];
  
         [traceAgent pHeader: [scnEntity getTC] value:"fulfilledOrder" fmt: Space];
         [traceAgent  pOrder: [scnEntity getTC] value: fulfilledOrder];

         currentPart->partQ = 0;
       }  //end  else of if ((currentOrder->itemQ > c
       [self accumuOutputInvCost: outPart]; // stats
     } //end if (currentPart->partQ != 0)
   } //end if ((currentPart = [self allocateStock:
  } // end if ((currentOrder->wantDate <= getCurrentTime())
 } // end while ((currentOrder = [index next]))
 [index drop];

 return self;
}


//---------get a part from inParts list---
-getPart: (Part *) inPart {
 [inPartBuf addLast: inPart];
 return self;
}

//---get a part from outParts --
-(Part *) getTopOfPartList: (id) aList {
 Part * aPart;
 Part * thePart;
 id index;

 index = [aList begin: [self getZone]];
 if ((aPart = [index next])) {
   thePart = [aPart copy: [self getZone]];
   [index drop];
   return thePart;
 }
 else {
   [index drop];
   return aPart;
 }

}

//---get outParts List
-getOutPartList {
 return self->outParts;
}

//---get a receiver from receivers list --
-(IdNo *) getTopOfReceivers {
 IdNo * aReceiver;
 id index;
 IdNo * theReceiver;

 index = [self->receivers begin: [self getZone]];
 if ((aReceiver = [index next])) {
   theReceiver = [IdNo  create: [self getZone]];
   theReceiver->no = aReceiver->no;
   [index drop];
   return theReceiver;
 }
 else {
   [index drop];
   return aReceiver;
 }

}

//---move top of outParts to the end of sentParts --
-moveFromOutPartsToSentParts {
 Part *aPart;
 Part *sentPart;

 aPart = [self->outParts removeFirst];
 sentPart = [Part create: [self getZone]];
 sentPart->partId = aPart->partId;
 sentPart->partQ  = aPart->partQ;
 sentPart->partArrTime = aPart->partArrTime;
 [self->sentParts addLast: sentPart];

//remove the receiver also
 [self->receivers removeFirst];  

 return self;
}


//----------check inventory by input partId ----
-(int) checkInv: (int) aPartId {
 Part * stockPart;

 if ((stockPart = [self allocateStock: aPartId]))
   return stockPart->partQ; 
 else
   return 0;

}

//------ point to a part in inStock
-(Part *) allocateStock: (int) identification {
id index;
Part * stockPart;
int found = FALSE;
int offset = -1;

 index = [inStock begin: [self getZone]];
 while ((stockPart = [index next]) && (found == FALSE)) {
   offset++;
   if (stockPart->partId == identification) {
      found = TRUE;
   }
 }
 [index drop];
 
 if (found == TRUE)
   stockPart = [self->inStock atOffset: offset];

 return stockPart;
}

//------
-addInStock: (Plan *) workOrder {
 Part * stockPart;
	
 if ((stockPart = [self allocateStock: workOrder->prodId])) {
   stockPart->partQ += workOrder->prodQ;
   [self addToInvStat: workOrder];
 }
 else {
   [self addNewStock: workOrder];
   [self addToInvStat: workOrder];
 }

 [self reduceCompDemandList: workOrder->prodId size: workOrder->prodQ];

 return self;  
}

//----initialize the initial stock according to EOQ--(MTS)
-initInStock {
id <List> eoqList;
Part * anEOQ;
id index;
Part * aPart;

 eoqList = [reOrdAgent getEoqList];
 index = [eoqList begin: [self getZone]];
 while ((anEOQ = [index next])) {
   [self->inStock addLast: anEOQ];
   aPart = [anEOQ copy: [self getZone]];
   aPart->partArrTime = 0.0;
   [self accumuInputInvCost: aPart];
 }
 [index drop];

 return self;

}

//------add new parts to inStock
-addNewStock: (Plan *) workOrder {
 Part * aPart;
 
 aPart = [Part create: [self getZone]];
 aPart->partId = workOrder->prodId;
 aPart->partQ  = workOrder->prodQ;
 [self->inStock addLast: aPart];

 return self;
}

//----
-(int) checkAvailComps: (Plan *) workOrder {
 id <List> compList;
 id index;
 IdNo * aComp;
 Part * compStock;
 int min = workOrder->prodQ;

 compList = [ofpModel getCompList: workOrder->prodId]; 
 index = [compList begin: [self getZone]];
 while ((aComp = [index next])) {
   if ((compStock = [self allocateStock: aComp->no])) {
      if (compStock->partQ < min)
        min = compStock->partQ;
   }
   else
     min = 0;
 }
 [index drop];

 return min;
}

//-- reduce the entry of compDemandList while receiving aPart
-reduceCompDemandList: (int) partNo size: (int) partQ {
 Order * anOrder;
 id index;
 int Qleft = partQ;
 Order * fulOrder;

 index = [self->compDemandList begin:[self getZone]];
 while ((anOrder = [index next]) && (Qleft > 0)) {
   if (anOrder->itemId == partNo) {
     if (Qleft >= anOrder->itemQ) {
       Qleft -= anOrder->itemQ;
       fulOrder = [anOrder copy: [self getZone]];
       [self->compFulfilledOrders addLast: fulOrder];
       [index remove];
     }
     else {
       anOrder->itemQ -= Qleft;
       fulOrder = [anOrder copy: [self getZone]];
       fulOrder->itemQ = Qleft;
       [self->compFulfilledOrders addLast: fulOrder];
       Qleft = 0;
     }
   }
 }
 [index drop];

 return self;
}

//---- remove components from inStock to factory
-removeToDispatch: (Plan *) workOrder {
 id <List> compList;
 id index;
 IdNo * aComp;
 Part * compStock;
 Part * statPart;

 compList = [ofpModel getCompList: workOrder->prodId]; 
 index = [compList begin: [self getZone]];
 while ((aComp = [index next])) {
   if ((compStock = [self allocateStock: aComp->no])) {
     compStock->partQ -= workOrder->prodQ;
     // statistics
     statPart = [Part create: [self getZone]];
     statPart->partId = aComp->no;
     statPart->partQ  = workOrder->prodQ;
     statPart->partArrTime = getCurrentTime();  
     [self accumuOutputInvCost: statPart]; // stats
   }
 }
 [index drop];

 return self;
}

//------final inventory in stock at the statistics time
-finalInv {
 Part * aPart;
 id index;

 index = [self->inStock begin:[self getZone]];
 while ((aPart = [index next])) {
   if (aPart->partQ > 0) {
     aPart->partArrTime = getCurrentTime();
     [self accumuOutputInvCost: aPart]; // stats
   }
 }

 return self;
}

//------add an compOrder to compDemandList --
-addToCompDemandList: (Order *) anOrder {

 [self->compDemandList addLast: anOrder];
 return self;
}

//------ add an order into demandList --
-addToDemandList: (Order *) demandOrder {
 id index;
 Order * anOrder;
 int found = FALSE;

 [traceAgent pHeader: [scnEntity getTC] value:"demandOrder" fmt: Space];
 [traceAgent pOrder: [scnEntity getTC] value: demandOrder];

 index = [self->demandList begin: [self getZone]];
 while ((anOrder = [index next]) && (found == FALSE)) {
   if (anOrder->dueDate > demandOrder->dueDate) {
     [index addBefore: demandOrder];
     found = TRUE;
   }
 }
 [index drop];

 if (found == FALSE) {
   [self->demandList addLast: demandOrder];
 }
 
 index = [self->demandList begin: [self getZone]];
 while ((anOrder = [index next])) {
 
 [traceAgent pHeader: [scnEntity getTC] value: "demandList" fmt: Space];
 [traceAgent pOrder: [scnEntity getTC] value: anOrder];
    
 }
 [index drop];
 
 return self;
}

//---------accumulate the number of product itemId on orderList---
-(int) sumOfDemandList: (Order *) order {
int total = 0;
id index;
Order * anOrder;

   index = [self->demandList begin: [self getZone]];
   while ((anOrder = [index next])) {
     if (anOrder->itemId == order->itemId) 
      total += anOrder->itemQ;
   }
   [index drop];

   return total;
}
-(int) sumOfCompDemandList: (Order *) order {
int total = 0;
id index;
Order * anOrder;

   index = [self->compDemandList begin: [self getZone]];
   while ((anOrder = [index next])) {
     if (anOrder->itemId == order->itemId) 
      total += anOrder->itemQ;
   }
   [index drop];

   return total;
}

//-----get the orders from demandList between start to end times--
-getPreDemand: (id) prodList from: (float) startTime to: (float) endTime source: (id) demandProds {
 id <List> preDemands;
 id index1, index2;
 Order * anOrder, * theOrder;
 int found = FALSE;
 float needTime;

 preDemands = [List create: [self getZone]];
 if ([demandProds getCount] <= 0)
   return preDemands;

 index1 = [demandProds begin: [self getZone]];
 while ((anOrder = [index1 next])) {
   if (anOrder->wantDate > 0) 
     needTime = anOrder->wantDate;
   else if (anOrder->dueDate > 0)
     needTime = anOrder->dueDate;
   else
     needTime = -1.0;
   
   if ((needTime > startTime) && (needTime <= endTime)) {
     if ([self isInInvSet: anOrder->itemId in: prodList] == Yes) {
       index2 = [preDemands begin: [self getZone]];
       while ((theOrder = [index2 next]) && (found == FALSE)) {
         if (theOrder->itemId == anOrder->itemId) {
           theOrder->itemQ += anOrder->itemId;
           found = TRUE;
         }
       }
       [index2 drop];
       if (found == FALSE) {
         theOrder = [Order create: [self getZone]];
         theOrder->itemId = anOrder->itemId;
         theOrder->itemQ = anOrder->itemQ;
         [preDemands addLast: theOrder];
       }
     }
   }
 }    
 [index1 drop];

 return preDemands;
}

// check if prodNo in the prodList --
-(int) isInInvSet: (int) prodNo in: (id) prodList {
int belonging;
id index;
IdNo * aProd;
  
  index = [prodList begin: [self getZone]];
  belonging = FALSE;
  while ((aProd = [index next]) && (belonging == FALSE)) {
     if (aProd->no == prodNo)
       belonging = TRUE;
  }

 return belonging;
}

-getDemandList {
 return self->demandList;
}

-getCompDemandList {
 return self->compDemandList;
}

-getFulfilledOrders {
 return self->fulfilledOrders;
}

-getCompFulfilledOrders {
 return self->compFulfilledOrders;
}

//sssssss------STATISTICS --
//-----order cycle time
-computeOrdCycleTime: (Order *) demandOrder {
 Ttime * aCycle;

 aCycle = [Ttime create: [self getZone]];
 aCycle->ttime = getCurrentTime() - demandOrder->ordArrTime;
 [self->ordCycleTs addLast: aCycle];
 
 [traceAgent pHeader: [scnEntity getTC] value: "cycleTime" fmt: Space];
 [traceAgent pFloat: [scnEntity getTC] value: aCycle->ttime fmt:NL];
 
 return self;
}

// average order cyele time
-(float) getAvgCycleT {
 float avgT;
 float totalTime = 0.0;
 Ttime * aCycle;
 id index;
 int counter = 0;
 int found = FALSE;

 index = [self->ordCycleTs begin: [self getZone]];
 while ((aCycle = [index next])) {
   totalTime += aCycle->ttime;
   counter++;
   found = TRUE;
 }
 [index drop];

 if (found == TRUE)
   avgT = totalTime / counter;
 else 
   avgT = -1.0;
 
 [traceAgent pHeader: [scnEntity getTC] value: "#ofOrders" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: counter fmt:NL];
 

 return avgT;
}

//---- calcurate fulfilledRate ---
-(float) getFulfilledRate {
id index;
Order * anOrder;
float fulfilled = 0.0;
float  notFulfilled = 0.0;
float fulfilledRate;

 index = [self->demandList begin:[self getZone]];
 while ((anOrder = [index next])) {
   notFulfilled += anOrder->itemQ;
 }
 [index drop];

 index = [self->fulfilledOrders begin:[self getZone]];
 while ((anOrder = [index next])) {
   fulfilled += anOrder->itemQ;
 }
 [index drop];

 [traceAgent pHeader: [scnEntity getTC] value: "SCN# " fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getScnId] fmt:NL];
 
 [traceAgent pHeader: [scnEntity getTC] value: "fulfilledTotal" fmt: Space];
 [traceAgent pFloat: [scnEntity getTC] value: fulfilled fmt:NL];
 
 [traceAgent pHeader: [scnEntity getTC] value: "notFulfilledTotal" fmt: Space];
 [traceAgent pFloat: [scnEntity getTC] value: notFulfilled fmt:NL];
 
 if ((fulfilled + notFulfilled) > 0.0) 
   fulfilledRate = (fulfilled / (fulfilled + notFulfilled));
 else
   fulfilledRate = -1.0;

 return fulfilledRate;
}

//----- inventory costs
-accumuInputInvCost: (Part *) inPart {
 id index;
 int found = FALSE;
 INV * aCost;
 float unitCost;

 unitCost = [scnEntity  getInvCost: inPart->partId];
 index = [self->accuInInvCost begin:[self getZone]];
 while ((aCost = [index next]) && (found == FALSE)) {
   if (aCost->prodId == inPart->partId) {
     aCost->prodQ  += inPart->partQ; 
     aCost->cost   += (inPart->partQ * inPart->partArrTime * unitCost);
     found = TRUE;
     
     [traceAgent pHeader: [scnEntity getTC] value: "inCost=" fmt: Space];
     [traceAgent pINVCost: [scnEntity getTC] value: aCost];

   }
 }
 if (found == FALSE) {
   aCost = [INV create: [self getZone]];
   aCost->prodId = inPart->partId;
   aCost->prodQ  = inPart->partQ;
   aCost->cost   = (inPart->partQ * inPart->partArrTime * unitCost);
   [self->accuInInvCost addLast: aCost];

   [traceAgent pHeader: [scnEntity getTC] value: "inCost=" fmt: Space];
   [traceAgent pINVCost: [scnEntity getTC] value: aCost];

 }
 [index drop];
 
 return self;
}

//--- Statistics for inventory
-addToInvStat: (Plan *) workOrder {
  Part * statPart;

 statPart = [Part create: [self getZone]];
 statPart->partId = workOrder->prodId;
 statPart->partQ = workOrder->prodQ;
 statPart->partArrTime = getCurrentTime();  
 [self accumuInputInvCost: statPart]; // stats

 return self;
}

// accumulate the output inventory cost
-accumuOutputInvCost: (Part *) outPart {
 id index;
 int found = FALSE;
 INV * aCost;
 float unitCost;

 unitCost = [scnEntity  getInvCost: outPart->partId];
 index = [self->accuOutInvCost begin:[self getZone]];
 while ((aCost = [index next]) && (found == FALSE)) {
   if (aCost->prodId == outPart->partId) {
     aCost->prodQ += outPart->partQ;
     aCost->cost += (outPart->partQ * outPart->partArrTime * unitCost);
     found = TRUE;
     [traceAgent pHeader: [scnEntity getTC] value: "outCost=" fmt: Space];
     [traceAgent pINVCost: [scnEntity getTC] value: aCost];
   }
 }
 if (found == FALSE) {
   aCost = [INV create: [self getZone]];
   aCost->prodId = outPart->partId;
   aCost->prodQ  = outPart->partQ;
   aCost->cost   = (outPart->partQ * outPart->partArrTime * unitCost);
   [self->accuOutInvCost addLast: aCost];
  
   [traceAgent pHeader: [scnEntity getTC] value: "outCost=" fmt: Space];
   [traceAgent pINVCost: [scnEntity getTC] value: aCost];

 }
 [index drop];

 
 return self;
}

// Average inventory cost
-(float) getAvgInvCost {
 id inIndex, outIndex;
 INV *inCost, *outCost;
 double avgCost = 0.0;
 int found = FALSE;
 int gotten = FALSE;

 [self finalInv];           
 inIndex = [self->accuInInvCost begin:[self getZone]];
 while ((inCost = [inIndex next])) {
   gotten = TRUE;
   outIndex = [self->accuOutInvCost begin:[self getZone]];
   found = FALSE;
   while ((outCost = [outIndex next]) && (found == FALSE)) {
     if (outCost->prodId == inCost->prodId) {
       avgCost += (outCost->cost - inCost->cost);
       found = TRUE;
     }
   }
   [outIndex drop];
 }
 [inIndex drop];

 if (gotten == TRUE)
   return avgCost;
 else
   return -1.0;
}

//--------statistics for due date performance
-computeDueDatePerformance: (Order *) demandOrder{
 Ttime * tardyD, * earlyD, *tardyW, * earlyW;
 float floatV;

 // tardiness and earliness at duedate measure
 floatV = getCurrentTime() - demandOrder->dueDate;
 tardyD = [Ttime create: [self getZone]];
 earlyD = [Ttime create: [self getZone]];
 if (floatV >= 0.0) {
   tardyD->ttime = floatV;
   earlyD->ttime = 0.0;
 }
 else {
   tardyD->ttime = 0.0;
   earlyD->ttime = 0.0 - floatV;
 }
 [self->tardyDL addLast: tardyD]; 
 [self->earlyDL addLast: earlyD];

 // tardiness and earliness at wantDate measure
 floatV = getCurrentTime() - demandOrder->wantDate;
 tardyW = [Ttime create: [self getZone]];
 earlyW = [Ttime create: [self getZone]];
 if (floatV >= 0.0) {
   tardyW->ttime = floatV;
   earlyW->ttime = 0.0;
 }
 else {
   tardyW->ttime = 0.0;
   earlyW->ttime = 0.0 - floatV;
 }
 [self->tardyWL addLast: tardyW]; 
 [self->earlyWL addLast: earlyW];

 return self;
}

// average due date performance
-(float) getAvgDueDatePerformance: (id) dueDateList {
 float avgT;
 float totalTime = 0.0;
 Ttime * aTime;
 id index;
 int counter = 0;
 int found = FALSE;

 index = [dueDateList begin: [self getZone]];
 while ((aTime = [index next])) {
   totalTime += aTime->ttime;
   counter++;
   found = TRUE;
 }
 [index drop];

 if (found == TRUE)
   avgT = totalTime / counter;
 else 
   avgT = -1.0;

 return avgT;
}

//---- get due date performance lists
-getTardyDL {
 return self->tardyDL;
}
-getEarlyDL {
 return self->earlyDL;
}
-getTardyWL {
 return self->tardyWL;
}
-getEarlyWL {
 return self->earlyWL;
}

//----------set the agent which processes orders------
-setOrdAgent: (id) agent {
 ordAgent = agent;
 return self;
}

-setScnEntity: (id) agent {
 scnEntity = agent;
 return self;
}

-setOfpModel: (id) agent {
 ofpModel = agent;
 return self;
}

-setReOrdAgent: (id) agent {
  reOrdAgent = agent;
  return self;
}


-setTraceAgent: (id) agent {
 traceAgent = agent;
 return self;
}

@end




