// Copyright (C) 1996 Fu-Ren Lin
// MaterialPlanning.m

#import "MaterialPlanning.h"

@implementation MaterialPlanning
+create: (id) aZone {
 MaterialPlanning * newMrp;
 newMrp = [super  create: aZone];
 newMrp->mrpPlan = [List create: aZone];
 newMrp->processedPlan = [List create: aZone];
 
 return newMrp;
}


//---------material planning process--------
-process {
 Plan * aPlan;
 id index;
 id <List> compList;
 IdNo * aComp;
 int currentInv, inOrdering;
 int planNotDispatched;  //mrped plan but not dispatched
 int needToPurchase;
 Order * outOrder;
 int complete = FALSE;

 [traceAgent pHeader: [scnEntity getTC] value: "MaterialP=" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getScnId] fmt:NL];

 // check and get the buildplan which are ready for mrp
 if ((aPlan = [prdPlanningAgent checkBuildPlanForMrp])) {
   while (complete == FALSE) {
     [self->mrpPlan addLast: aPlan];
     [prdPlanningAgent markBuildPlanForMrp: DONE];
     if ((aPlan = [prdPlanningAgent checkBuildPlanForMrp]))
       complete = FALSE;
     else
       complete = TRUE;
   }  
 }

 // material allocation
 while ((aPlan = [self getTopOfMrpPlan])) {
   
   compList = [ofpModel getCompList: aPlan->prodId];
   index = [compList begin: [self getZone]];
   while ((aComp = [index next])) {
     planNotDispatched = [self checkCompProcessedPlanNotDispatched: aPlan->prodId comp: aComp->no];
     [traceAgent pHeader: [scnEntity getTC] value: "planNotDispatched =" fmt: Space];
     [traceAgent pInt: [scnEntity getTC] value: planNotDispatched fmt: NL];
currentInv = [invAgent checkInv: aComp->no];
     [traceAgent pHeader: [scnEntity getTC] value: "currentInv =" fmt: Space];
     [traceAgent pInt: [scnEntity getTC] value:  currentInv fmt: NL];
 
     inOrdering =  [ordAgent checkInOrderingForMrp: aComp];
     [traceAgent pHeader: [scnEntity getTC] value: "inOrdering =" fmt: Space];
     [traceAgent pInt: [scnEntity getTC] value: inOrdering fmt: NL];
     
     if ((currentInv + inOrdering - planNotDispatched) >= aPlan->prodQ)
       needToPurchase = 0;
     else
       needToPurchase = aPlan->prodQ-(currentInv+inOrdering-planNotDispatched);

     if (needToPurchase > 0) {  // purchase the components
       outOrder = [Order create: [self getZone]];
       outOrder->itemId = aComp->no;
       outOrder->itemQ  = needToPurchase;
       outOrder->ordArrTime = getCurrentTime();
       outOrder->wantDate = aPlan->scheduledStart;
       outOrder->dueDate  = aPlan->scheduledStart;
       outOrder->senderId  = [scnEntity getScnId];
       [ordAgent addOneToOutOrders: outOrder];
     }
   }
   [index drop];
   [self moveFromMrpPlanToProcessedPlan];
 }

 return self;
}

//SYN****** USED FOR SYNCHRONIZATION STRATEGY ---
//-- check the availability of material for an mfgOrder --
-(float) checkMatAvail: (Order *) mfgOrder level: (int) upProp {
 float availTime;
 id index;
 id <List> compList;
 IdNo * aComp;
 int currentInv, inOrdering;
 int planNotDispatched;  //mrped plan but not dispatched
 int needToPurchase;
 int avail, receiving;
 int aSupplierNo;
 SCNEntitySwarm * aSupplierE;
 Order * outOrder, * anOrder;
 float maximum = 0;
 float deliverTime;
 int upShift;

   [traceAgent pHeader: [scnEntity getTC] value: "mat:upProp=" fmt: Space];
   [traceAgent pInt: [scnEntity getTC] value: upProp fmt: NL];

 // determine the availability time of material for mfgOrder
  
   compList = [ofpModel getCompList: mfgOrder->itemId];
   index = [compList begin: [self getZone]];
   while ((aComp = [index next])) {
      planNotDispatched = [self checkCompProcessedPlanNotDispatched: mfgOrder->itemId comp: aComp->no];
     [traceAgent pHeader: [scnEntity getTC] value: "planNotDispatched =" fmt: Space];
     [traceAgent pInt: [scnEntity getTC] value: planNotDispatched fmt: NL];
 currentInv = [invAgent checkInv: aComp->no];
     [traceAgent pHeader: [scnEntity getTC] value: "currentInv =" fmt: Space];
     [traceAgent pInt: [scnEntity getTC] value:  currentInv fmt: NL];
 
     inOrdering =  [ordAgent checkInOrderingForMrp: aComp];
     [traceAgent pHeader: [scnEntity getTC] value: "inOrdering =" fmt: Space];
     [traceAgent pInt: [scnEntity getTC] value: inOrdering fmt: NL];
     
     if ((currentInv + inOrdering - planNotDispatched) >= mfgOrder->itemQ)  
       needToPurchase = 0;
      else
       needToPurchase = mfgOrder->itemQ-(currentInv+inOrdering-planNotDispatched);

     // determine the available time of components
     if (needToPurchase > 0)
      avail = mfgOrder->itemQ - needToPurchase;
     else
      avail = mfgOrder->itemQ;
    
     //from inventory or receiving
     if (needToPurchase == 0) {  
       if ((currentInv - planNotDispatched) >= avail) { // all from inv
         availTime = mfgOrder->ordArrTime;
       }
       else if ((currentInv + inOrdering - planNotDispatched) >= avail) { // some from receiving
         receiving = avail - (currentInv - planNotDispatched);
         anOrder = [mfgOrder copy: [self getZone]];
         anOrder->itemId = aComp->no;
         anOrder->itemQ = receiving;
         anOrder->ordArrTime = mfgOrder->ordArrTime;
         availTime = [ordAgent estimateProdArrivalTime: anOrder level: upProp];
       }
     }
    
     // purchase the components
     if (needToPurchase > 0) {   
       if (upProp == 0)     // estimated lead time isolated
         availTime = mfgOrder->ordArrTime + [scnEntity getOrderDelayAndPartDeliverWithSupplier: aComp->no];
       else {                    // InfoShareUp
         outOrder = [mfgOrder copy: [self getZone]];
         outOrder->itemId = aComp->no;
         outOrder->itemQ  = needToPurchase;
         outOrder->senderId  = [scnEntity getScnId];
         if ([scnEntity getMultiSuppliers] == No) {  // MonoSupplier
           aSupplierNo = [scnEntity chooseSupplier: aComp->no];
           aSupplierE = [ofpModel allocateEntity: aSupplierNo];
           deliverTime = [aSupplierE getDeliverTime: [scnEntity getScnId]];
           outOrder->ordArrTime += [scnEntity getDelayTime: aSupplierNo];
           upShift = upProp-1;
           availTime = deliverTime + [aSupplierE acceptCommitDateQueryMessage: outOrder level: upShift];
           [traceAgent pHeader: [scnEntity getTC] value: "mat:avail-S=" fmt: Space];
           [traceAgent pFloat: [scnEntity getTC] value: availTime fmt: NL];
         }
         else if ([scnEntity getMultiSuppliers] == Yes) { // MultiSupplier
           upShift = upProp - 1;
           availTime = [self determineCommitDateFromMultiSuppliers: outOrder level: upShift]; 
           [traceAgent pHeader: [scnEntity getTC] value: "mat:avail-M=" fmt: Space];
           [traceAgent pFloat: [scnEntity getTC] value: availTime fmt: NL];
         }
        }
      }
    if (availTime > maximum)  // the latest avail of a componet as availTime.
      maximum = availTime;
  }
  [index drop];
 
 return maximum;
}

//----get top of mrpPlan
-(Plan *)getTopOfMrpPlan {
 id index;
 Plan * aPlan;

 index = [self->mrpPlan begin: [self getZone]];
 aPlan = [index next];
 [index drop];
 return aPlan;
}

//-----move top of mrpPlan to the bottom of processedPlan
-moveFromMrpPlanToProcessedPlan {
 id index;
 Plan * thePlan, * aPlan;
 Plan * pWkOrder;

 aPlan = [self->mrpPlan removeFirst];
 thePlan = [aPlan copy: [self getZone]];
 [self->processedPlan addLast: thePlan];

 index = [self->processedPlan begin: [self getZone]];
 while ((pWkOrder = [index next])) {
   [traceAgent pHeader: [scnEntity getTC] value: "processedPlan" fmt: Space];
   [traceAgent pPlan: [scnEntity getTC] value: pWkOrder];
  }
 [index drop];

 return self;
}

//-------check the processedPlan which is not dispatched (on product level)
-(int) checkProcessedPlanNotDispatched: (int) prodNo {
  id index;
  Plan * pWkOrder;
  int num = 0;

  index = [self->processedPlan begin: [self getZone]];
  while ((pWkOrder = [index next])) {
    if (pWkOrder->prodId == prodNo)
      num += pWkOrder->prodQ;
  }
  [index drop];
  
  return num;
}
//-------check the processedPlan which is not dispatched (including checking the product having the shared common components)
-(int) checkCompProcessedPlanNotDispatched: (int) prodNo comp: (int) compNo {
  id index1, index2;
  Plan * pWkOrder;
  int num = 0;
  id <List> prodList;
  IdNo * aProd;
  id <List> compList;
  id <List> otherCompList;
  int theProdNo;

  compList = [ofpModel getCompList: prodNo];

  prodList = [scnEntity getProdIds];
  index1 = [prodList begin: [self getZone]];
  while ((aProd = [index1 next])) {
    theProdNo = -1;
    if (aProd->no == prodNo) 
      theProdNo = prodNo;
    else {
      otherCompList = [ofpModel getCompList: aProd->no];
      if ([self hasCommonComp:compNo in:otherCompList] == Yes) 
        theProdNo = aProd->no;
    }
    if (theProdNo > 0) {   
      index2 = [self->processedPlan begin: [self getZone]];
      while ((pWkOrder = [index2 next])) {
        if (pWkOrder->prodId == theProdNo)
          num += pWkOrder->prodQ;
      }
      [index2 drop];
    }
  }
  [index1 drop];

  return num;
}


//--compare two lists to find out any common entries --
-(int) hasCommonComp: (int) aCompNo in: (id) bList {
 int hasCommon = No;
 id index;
 IdNo * bComp;

 index = [bList begin: [self getZone]];
 while ((bComp = [index next])) {
   if (aCompNo == bComp->no)
       hasCommon = Yes;
 }
 [index drop];

 return hasCommon;
}

//-------reduce the processedPlan whose prod is dispatched
-removeProcessedPlan: (Plan *) aPlan {
  id index;
  Plan * pWkOrder;
  int Qleft = aPlan->prodQ;

  index = [self->processedPlan begin: [self getZone]];
  while ((pWkOrder = [index next]) && (Qleft > 0)) {
    [traceAgent pHeader: [scnEntity getTC] value: "processedPlan to dispatch" fmt: Space];
    [traceAgent pPlan: [scnEntity getTC] value: pWkOrder];

    if (pWkOrder->prodId == aPlan->prodId) {
      if (Qleft < pWkOrder->prodQ) {
        pWkOrder->prodQ -= Qleft;
        Qleft = 0;
      }
      else {
        Qleft -= pWkOrder->prodQ;
        [index remove];
      }
    }
  }
  [index drop];

  return self;
}


//--compute the latest avail time of a "comp" from suppliers --
-(float) determineCommitDateFromMultiSuppliers: (Order *) outOrder level:
(int) upProp { 
id <List> theSupplierList;
IdNo * aSupplierId;
SCNEntitySwarm * aSupplierE;
float availTime;
float minimum = Infinite;
int found = FALSE;
id index;
float deliverTime, commitDate;
Order *anOrder;

 theSupplierList = [scnEntity chooseMultiSuppliers: outOrder->itemId];
 index = [theSupplierList begin:[self getZone]];
 while ((aSupplierId = [index next])) {
   aSupplierE = [ofpModel allocateEntity: aSupplierId->no];
   deliverTime = [aSupplierE getDeliverTime: [scnEntity getScnId]];
   if (upProp == 0) {
     availTime = [scnEntity getDelayTime: aSupplierId->no];
     availTime += outOrder->ordArrTime;
   }
   else {
     anOrder = [outOrder copy: [self getZone]];
     anOrder->ordArrTime += [scnEntity getDelayTime: aSupplierId->no];
     availTime = [aSupplierE acceptCommitDateQueryMessage: anOrder level: upProp];
   }
   found = TRUE;
   commitDate = deliverTime + availTime;
  if (commitDate < minimum)
     minimum = commitDate;
 
   [traceAgent pHeader: [scnEntity getTC] value: "supplierNo =" fmt: Space];
   [traceAgent pInt: [scnEntity getTC] value: aSupplierId->no fmt: Tab];
   [traceAgent pHeader: [scnEntity getTC] value: "commitDate =" fmt: Space];
   [traceAgent pFloat: [scnEntity getTC] value: commitDate fmt: NL];
 } 
 [index drop];

 if (found == FALSE) 
   commitDate = Infinite;
 else
   commitDate = minimum;

 return commitDate;
}


//----estimate lead time-----------
-(float) getLeadTime:  (int) prodId {
 id <List> compList;
 IdNo * aComp;
 id index;
 float maxTime = 0.0;
 float deliverT;

 compList = [ofpModel getCompList: prodId];
 index = [compList begin: [self getZone]];
 while ((aComp = [index next])) {
   deliverT = [scnEntity getOrderDelayAndPartDeliverWithSupplier: aComp->no];
   if (maxTime < deliverT) 
     maxTime = deliverT;
 }
 [index drop];
 return maxTime;
}


//------set its scnEntity ----
-setScnEntity: (id) agent {
 scnEntity = agent;
 return self;
}

-setOrdAgent: (id) agent {
 ordAgent = agent;
 return self;
}

-setInvAgent: (id) agent {
 invAgent = agent;
 return self;
}

-setPrdPlanningAgent: (id) agent {
 prdPlanningAgent = agent;
 return self;
}

-setOfpModel: (id) agent {
 ofpModel = agent;
 return self;
}


-setTraceAgent: (id) agent {
 traceAgent = agent;
 return self;
}

@end

