//Copyright (C) 1996 Fu-ren Lin
//OFPModelSwarm.m

#import "OFPModelSwarm.h"

@implementation OFPModelSwarm
+create: (id) aZone {
  OFPModelSwarm * newOfpM;

  newOfpM = [super create: aZone];
  return newOfpM;
}

-buildObjects {
 [super buildObjects];

 // read the files for simulation
 [self getConfigFiles: configFile];

 // open debugfile for debugging and testing
 traceAgent = [Trace create: [self getZone]];
 [traceAgent openDebugFile];
 
 [self getScnConfig];
 [self initialization];

 return self;
}

//---------------
-buildActions {
  int i;
  SCNEntitySwarm * scnESwarm;

  [super buildActions];
 
// for debug purpose
// [self printForDebug];

// printClock schedule
  printClockSchedule = [Schedule createBegin: [self getZone]];
  [printClockSchedule setRepeatInterval: TimeStep];
  printClockSchedule = [printClockSchedule createEnd];
  [printClockSchedule at: 0 createActionTo: self message: M(printClock)];

  for (i=0; i<[scnEArray getCount]; i++) {
    scnESwarm = [scnEArray atOffset: i];
    [scnESwarm buildActions];
  }  
  return self;
}

//-----------------
-activateIn: (id) swarmContext {
  int i;
  SCNEntitySwarm * scnESwarm;

  [super activateIn: swarmContext];
  
  [printClockSchedule activateIn: self];

  for (i=0; i<[scnEArray getCount]; i++) {
    scnESwarm = [scnEArray atOffset: i];
    [scnESwarm activateIn: self];
  }  

  return [self getSwarmActivity];
}

//ffffff----obtain related files needed in simulation ---
-getConfigFiles: (char *) configFile {
 FILE   *fconfigfile, *fopen();
 char   junk[80];

if ((fconfigfile = fopen(configFile, "r")) == NULL) 
   [SourceMessage raiseEvent: "Error in opening file configFile\n"];

   strcpy(Notes, configFile); // for denoting the output report.
// obtain the file name of a SCN configuration file
  [self skipComments: fconfigfile];
  fscanf(fconfigfile, "%s %s", junk, config);

// obtain the file name of a SCN strategy file
  [self skipComments: fconfigfile];
  fscanf(fconfigfile, "%s %s", junk, strategyFile);

// obtain the name of input orders file
  [self skipComments: fconfigfile];
  fscanf(fconfigfile, "%s %s", junk, orderFile);

// obtain the name of material input file
  [self skipComments: fconfigfile];
  fscanf(fconfigfile, "%s %s", junk, partFile);

// obtain the name of debug file
  [self skipComments: fconfigfile];
  fscanf(fconfigfile, "%s %s", junk, debugFile);

// obtain the name of report file
 [self skipComments: fconfigfile];
 fscanf(fconfigfile, "%s %s", junk, reportFile);

// obtain the end simulation day
 [self skipComments: fconfigfile];
 fscanf(fconfigfile, "%s %f", junk, &endDay);
 
 fclose(fconfigfile);
 return self;
}


//xxxxx---- obtain parameters, BOM, SCN'strcture, policies --
-getScnConfig { 
 FILE *fconfig, *fopen();
 FILE *fstrategy, *fopen();
 int i, j; 
 char mfgCap;
 int count, intV;
 float floatV;
 IdNo * aComp;
  char junk[80];
 Components * comp;
 SCNEntitySwarm * scnESwarm;
 id <List> invTypes;
 IdNo * invType;

 if ((fconfig = fopen(config, "r")) == NULL) 
   [SourceMessage raiseEvent: "Error in opening file config\n"];
 
 if ((fstrategy = fopen(strategyFile, "r")) == NULL) 
   [SourceMessage raiseEvent: "Error in opening file config\n"];
 
 //Section A --obtain the global parameters
  [self skipComments: fconfig];
  fscanf(fconfig, "%s %d", junk, &PlanHorizone);
  
  [self skipComments: fconfig];
  fscanf(fconfig, "%s %f", junk, &UNIT);

    [self skipComments: fconfig];
  fscanf(fconfig, "%s %d", junk, &ForecastingH);

  [self skipComments: fconfig];
  fscanf(fconfig, "%s %f", junk, &Alpha);

  [self skipComments: fconfig];
  fscanf(fconfig, "%s %d", junk, &EOQPeriod);
 
  [self skipComments: fconfig];
  fscanf(fconfig, "%s %d", junk, &ROPCheckH);

  [self skipComments: fconfig];
  fscanf(fconfig, "%s %d", junk, &TimeStep);
  
//Section B -- obtain the product structure (compArray)
  [self skipComments: fconfig];
  fscanf(fconfig, "%s %d %s", junk, &NumOfParts, junk);

  self->compArray = [Array create: [self getZone] setCount: NumOfParts];

// compute the product constituent lists  
 for (i = 0; i< NumOfParts; i++) {
   comp = [Components create: [self getZone]]; 
   fscanf(fconfig, "%d", &intV);
   comp->prodId = intV;
   fscanf(fconfig, "%d", &intV);
   comp->compQ = intV;
   if (comp->compQ > 0) {
     comp->compList = [List create: [self getZone]];
     for (j=0; j<comp->compQ; j++) {
       aComp = [IdNo create: [self getZone]];
       fscanf(fconfig, "%d", &intV);
       aComp->no = intV;
       [comp->compList addLast: aComp];
     }
   }
   [self->compArray atOffset: i put: comp];
 }

//Section C--- define SCN structure ------
// obtain the number of scn entities.
  [self skipComments: fconfig];
  fscanf(fconfig, "%s %d", junk, &NumOfEntities);

  // create scnEArray 
  self->scnEArray = [Array create: [self getZone] setCount: NumOfEntities];
    
  // create each scnESwarm
  for (i=0; i<NumOfEntities; i++) {  
    scnESwarm = [SCNEntitySwarm create: [self getZone]];
    [scnEArray atOffset: i put: scnESwarm];
  }
  
  // setup properties  
  for (i=0; i<NumOfEntities; i++) {  
    scnESwarm = [scnEArray atOffset: i];
    [self skipComments: fconfig];
    fscanf(fconfig, "%s %c", junk, &mfgCap);
    [scnESwarm setMfgCap: mfgCap];
    [scnESwarm  setScnId: i];
    [scnESwarm  setOfpModel: self];

  // its suppliers
    [self skipComments: fconfig];
    fscanf(fconfig, "%s %d", junk, &count);
    for (j=0; j<count; j++) {
      fscanf(fconfig, "%d", &intV);
      [scnESwarm addOneSupplier: intV];
    }
  // its product order cost according to the suppliers
    [self skipComments: fconfig];
    fscanf(fconfig, "%s %d", junk, &count);
    for (j=0; j<count; j++) {
      fscanf(fconfig, "%f", &floatV);
      [scnESwarm addOneOrderCost: floatV at:j];
    }
  // its customers
    [self skipComments: fconfig];
    fscanf(fconfig, "%s %d", junk, &count);
    for (j=0; j<count; j++) {
      fscanf(fconfig, "%d", &intV);
      [scnESwarm addOneCustomer: intV];
    }

  // its order delay time
    [self skipComments: fconfig];
    fscanf(fconfig, "%s %d", junk, &count);
    for (j=0; j<count; j++) {
      fscanf(fconfig, "%f", &floatV);
      [scnESwarm addOneOrderDelayT: floatV];
    }

  // its part deliver time
    [self skipComments: fconfig];
    fscanf(fconfig, "%s %d", junk, &count);
    for (j=0; j<count; j++) {
      fscanf(fconfig, "%f", &floatV);
      [scnESwarm addOnePartDeliverT: floatV];
    }
  // its inventory unit cost according to products
   invTypes = [List create: [self getZone]];
   [self skipComments: fconfig];
   fscanf(fconfig, "%s %d", junk, &count);
   for (j=0; j<count; j++) {
     fscanf(fconfig, "%d", &intV);
     invType = [IdNo create: [self getZone]];
     invType->no = intV;
     [invTypes addLast: invType];
   }
   
   [self skipComments: fconfig];
   fscanf(fconfig, "%s %d", junk, &count);
   for (j=0; j<count; j++) {
     fscanf(fconfig, "%f", &floatV);
     [scnESwarm addOneInvCost: floatV types:invTypes at: j];
   }

  // its product list and their processing time for entities with mfgCap
    if ([scnESwarm getMfgCap] == 'Y') {
      [self skipComments: fconfig];
      fscanf(fconfig, "%s %d", junk, &count);
      for (j=0; j<count; j++) {
        fscanf(fconfig, "%d", &intV);
        [scnESwarm addOneProdId: intV];      //product group
      }
      
      [self skipComments: fconfig];
      fscanf(fconfig, "%s %d", junk, &count);
      for (j=0; j<count; j++) {
        fscanf(fconfig, "%f", &floatV);
        [scnESwarm addOneMfgProcT: floatV];   // mfg process time
      } 
    } // end of if (scnESwarm->withMfgCap ...

    // obtain the inventory quantity
    [self skipComments: fconfig];
    fscanf(fconfig, "%s %d", junk, &count);
    for (j=0; j<count; j++) {
      fscanf(fconfig, "%d", &intV);
      [scnESwarm addOneProdDQ: intV at: j];         }

    // due date tightness 
    [self skipComments: fconfig];
    fscanf(fconfig, "%s %f", junk, &floatV);
    [scnESwarm assignDueDateTightness: floatV];  
   
// Multiple suppliers definition
    [self skipComments: fconfig];
    fscanf(fconfig, "%s %d", junk, &intV);
    [scnESwarm assignMultiSuppliers: intV];
 
    // trace options
    [self skipComments: fconfig];
    fscanf(fconfig, "%s %d", junk, &intV);
    [scnESwarm setTraceCode: intV];

  } // end of  for (i=0; i<NumOfEntities; i++)
  
//Section D ----assign the demand management policies -- (from other file)
  for (i=0; i<NumOfEntities; i++) {
    [self skipComments: fstrategy];
    fscanf(fstrategy, "%s %d", junk, &intV);
    scnESwarm = [self->scnEArray atOffset: i];
    [scnESwarm assignDemandPolicy: intV];
  }

//Section E----assign the coordination strategies -- (from other file)
  for (i=0; i<NumOfEntities; i++) {
    [self skipComments: fstrategy];
    fscanf(fstrategy, "%s %d", junk, &intV);
    scnESwarm = [self->scnEArray atOffset: i];
    [scnESwarm assignCoordIU: intV];

    fscanf(fstrategy, "%d", &intV);
    scnESwarm = [self->scnEArray atOffset: i];
    [scnESwarm assignCoordID: intV];

    fscanf(fstrategy, "%d", &intV);
    scnESwarm = [self->scnEArray atOffset: i];
    [scnESwarm assignCoordSMC: intV];

    fscanf(fstrategy, "%d", &intV);
    scnESwarm = [self->scnEArray atOffset: i];
    [scnESwarm assignCoordDBC: intV];

    fscanf(fstrategy, " %d", &intV);
    scnESwarm = [self->scnEArray atOffset: i];
    [scnESwarm assignUpPropLayer: intV];

    fscanf(fstrategy, "%d", &intV);
    scnESwarm = [self->scnEArray atOffset: i];
    [scnESwarm assignDownPropLayer: intV];
  }

//xxxxx---- build internal agents --
  for (i=0; i<NumOfEntities; i++) {  
    scnESwarm = [self->scnEArray atOffset: i];
    [scnESwarm buildObjects];
  }


  fclose(fconfig);
  fclose(fstrategy);

  return self;
}

//--------skip the comments in the input scn.conf and scn.files files
-skipComments: (FILE *) filePoint {
 char ch;
 
  do {
    fscanf(filePoint, "%c", &ch);
  } while (ch != '@');

  return self;
}

-printForDebug {
 int i;
 SCNEntitySwarm * scnESwarm;

 for (i=0; i<NumOfEntities; i++) {  
    scnESwarm = [scnEArray atOffset: i];
    [scnESwarm printSuppliers];
    [scnESwarm printCustomers];
 }
 
 return self;
}


//iiiii---initialization: input outside orders and materials into the model --
-initialization {
  int itemId, itemQ, receiver, partId, partQ;
  float ordArrTime, wantDate, dueDate, partArrTime;
  Order * order;
  Part  * part;
  FILE *forder, *fopen();
  FILE *fpart, *fopen();
  SCNEntitySwarm * scnESwarm;

// input orders assignment
  if ((forder = fopen(orderFile, "r")) == NULL) { 
    [SourceMessage raiseEvent: "Error in opening file orderFile"];
  }

  fscanf(forder, "%d %d %f %f %f %d\n", &itemId, &itemQ, &ordArrTime, &wantDate, &dueDate, &receiver);
  while ((ordArrTime <= (endDay * UNIT)) && (itemId >= 0)){ 
    order = [Order create: [self getZone]];
    order->itemId = itemId;
    order->itemQ  = itemQ;
    order->ordArrTime = ordArrTime;
    order->wantDate = wantDate;
    order->dueDate = dueDate;
    order->senderId = -1;
    scnESwarm = [scnEArray atOffset: receiver];
    [scnESwarm addOneOrderToInOrders: order];  
    fscanf(forder, "%d %d %f %f %f %d\n", &itemId, &itemQ, &ordArrTime,&wantDate, &dueDate, &receiver);	
   }
   fclose(forder);

// input parts assignment
  if ((fpart = fopen(partFile, "r")) == NULL) { 
    [SourceMessage raiseEvent: "Error in opening file partFile \n"];
  }
  
  fscanf(fpart, "%d %d %f %d\n", &partId,  &partQ, &partArrTime, &receiver);
  while ((partArrTime <= (endDay*UNIT)) && (partId >= 0)) {
    part = [Part create: [self getZone]];
    part->partId = partId;
    part->partQ  = partQ;
    part->partArrTime = partArrTime;
    scnESwarm = [scnEArray atOffset: receiver];
    [scnESwarm addOnePartToInParts: part];     
    fscanf(fpart, "%d %d %f %d\n", &partId,  &partQ, &partArrTime, &receiver);
  }
  fclose(fpart);
 
  return self;
}

//-----find the list of components for a product ---
-getCompList: (int) prodId  {
 id <List> compL;
 Components * comps;
 id index;
 int found = FALSE;

 index = [self->compArray begin: [self getZone]];
 while ((comps = [index next]) && (found == FALSE)) {
   if (comps->prodId == prodId) {
     compL = [comps->compList copy: [self getZone]];
     found = TRUE;
   }
 }
 [index drop];

 return compL;
}

//------Allocate the scn entity by given scnId --
-allocateEntity: (int) anId {
 SCNEntitySwarm * scnE;
 
 scnE = [scnEArray atOffset: anId];

 return scnE;
}
 

// get the scnEArray for OFPStatSwarm
-getScnEArray {
 return self->scnEArray;
}


-printClock {
 double cTime;

 cTime = getCurrentTime();
 [traceAgent pHeader: 2 value: ">>clock --" fmt: Space];
 [traceAgent pFloat: 2 value: cTime fmt:NL];
// [traceAgent pHeader: 1 value: ">>clock --" fmt: Space];
// [traceAgent pFloat: 1 value: cTime fmt:NL];
 return self;
}

@end

