// Copyright (C) 1996 Fu-ren Lin

#import "OFPObserverSwarm.h"
#import "OFPModelSwarm.h"
#import <collections.h>
#import <swarmobject.h>

@implementation OFPObserverSwarm

// createBegin: here we set up the default observation parameters.
+createBegin: (id) aZone {
  OFPObserverSwarm * obj;

  // Superclass createBegin to allocate ourselves.
  obj = [super createBegin: aZone];

  // Fill in the relevant parameters (only one, in this case).
  obj->displayFrequency = 1;

  return obj;
}

//-----------
// createEnd: create objects we know we'll need. 
-createEnd {
  return [super createEnd];
}

//------------
// Create the objects used in the display of the model. 
-buildObjects {
  id modelZone;					  // zone for model.

  [super buildObjects];
  
  // First, we create the model that we're actually observing. The
  // model is a subswarm of the observer. We also create the model in
  // its own zone, so storage is segregated.
  modelZone = [Zone create: [self getZone]];
  ofpModelSwarm = [OFPModelSwarm create: modelZone];
 
  [ofpModelSwarm buildObjects];
 
  statusM = [StatusM createBegin: [self getZone]];
  statusM = [statusM createEnd];

  statusI = [StatusI createBegin: [self getZone]];
  statusI = [statusI createEnd];
  
  // create graph for Material flow
  mflowGraph = [BLTGraph create: globalZone];
  [mflowGraph title: "Matrial Flow vs. Time"];
  [mflowGraph axisLabelsX: "time" Y: "Material Flow"];
  [mflowGraph setWidth: 400 Height: 250];

  
  // create three elements to draw on that graph.
  materialElement = [mflowGraph createElement];
  [materialElement setLabel: "Shipment of Products"];
  [materialElement setColor: "red"];
  [materialElement setWidth: 3];
  
  mflowGrapher = [ActiveGraph createBegin: globalZone];
  [mflowGrapher setElement: materialElement];
  [mflowGrapher setDataFeed: statusM]; 
  [mflowGrapher setProbedSelector: M(getM)]; 
  mflowGrapher = [mflowGrapher createEnd];
  [mflowGraph pack];
    
//Information flow

  iflowGraph = [BLTGraph create: globalZone];
  [iflowGraph title: "Information Flow vs. Time"];
  [iflowGraph axisLabelsX: "time" Y: "Information Flow"];
  [iflowGraph setWidth: 400 Height: 250];

  // create elements to draw on that graph.
  infoElement = [iflowGraph createElement];
  [infoElement setLabel: "Arrival of Orders"];
  [infoElement setColor: "green"];
  [infoElement setWidth: 3];
  
  iflowGrapher = [ActiveGraph createBegin: globalZone];
  [iflowGrapher setElement: infoElement];
  [iflowGrapher setDataFeed: statusI]; 
  [iflowGrapher setProbedSelector: M(getI)];
  iflowGrapher = [iflowGrapher createEnd];
  [iflowGraph pack];
  
  // All done - we're ready to build a schedule and go.
  return self;
}  

//-------------
// Create the actions necessary for the simulation. This is where
// the schedule is built (but not run!)
-buildActions {
  [super buildActions];
  
  // First, let our model swarm build its own schedule.
  [ofpModelSwarm buildActions];
  
  // Create an ActionGroup for display: a bunch of things that occur in
  // a specific order, but at one step of simulation time. 

   displayActions = [ActionGroup create: [self getZone]];
   [displayActions  createActionTo: mflowGrapher  message: M(step)];
   [displayActions  createActionTo: iflowGrapher  message: M(step)];
   [displayActions  createActionTo: controlPanel  message: M(doTkEvents)];

  // And the display schedule.
  displaySchedule = [Schedule createBegin: [self getZone]];
  [displaySchedule setRepeatInterval: displayFrequency]; // note frequency!
  displaySchedule = [displaySchedule createEnd];
  [displaySchedule at: 0 createAction: displayActions];
  
  return self;
}  

// activateIn: - activate the schedules so they're ready to run.
// The swarmContext argument has to do with what we were activated *in*.
// Typically the ObserverSwarm is the top-level Swarm, so it's activated
// in "nil". But other Swarms and Schedules and such will be activated
// inside of us.
-activateIn: (id) swarmContext {
  // First, activate ourselves (just pass along the context).
  [super activateIn: swarmContext];

  // Activate the model swarm in ourselves. The model swarm is a
  // subswarm of the observer swarm.
  [ofpModelSwarm activateIn: self];

  // Now activate our schedule in ourselves. This arranges for the
  // execution of the schedule we built.
  [displaySchedule activateIn: self];
  
  // Activate returns the swarm activity - the thing that's ready to run.
  return [self getSwarmActivity];
}

@end
