// Copyright (C) 1996 Fu-Ren Lin
// OrderManagement.m
// OrderManagement is a logical agent handling orders.

#import "OrderManagement.h"

@implementation OrderManagement
+create: (id) aZone {
 OrderManagement *newOrdM;
 newOrdM = [super create: aZone];
 newOrdM->outOrders = [List create: aZone];
 newOrdM->sentOrders = [List create: aZone];
 newOrdM->backOrders = [List create: aZone];
 newOrdM->inOrdBuf = [List create: aZone];
 return newOrdM;
}

//----------processing the incoming orders-------------
-process {
 Order * currentOrder, *outOrder, *anOrder;
 int needToOrder;
 Order * backOrder;

 [traceAgent pHeader: [scnEntity getTC] value: "ordM=" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getScnId] fmt:NL];

 // receive orders from inOrdBuf list
 while ((currentOrder = [self getTopOfOrderList: self->inOrdBuf])) {
   needToOrder = [self checkAvailInv: currentOrder]; 
   anOrder = [currentOrder copy: [self getZone]];  //in Public.m
   currentOrder->dueDate = [self estimateCommitDate: anOrder level: [scnEntity getUpPropLayer]];
   [invAgent addToDemandList: currentOrder];
  
   // need to re-order
   if (needToOrder > 0) {
     outOrder = [currentOrder copy: [self getZone]];
     outOrder->itemQ  = needToOrder;
     outOrder->ordArrTime = getCurrentTime();
     outOrder->senderId  = [scnEntity getScnId];
     if ([scnEntity getDemandPolicy] == MTS) {
       backOrder = [outOrder copy: [self getZone]];
       [self->backOrders addLast: backOrder];
     }
     if (([scnEntity getMfgCap] == 'N') ||    // re-ordering
           ([scnEntity getMfgCap] == 'S'))      // S for part source
        [self->outOrders addLast: outOrder];
     else if ([scnEntity getMfgCap] == 'Y') //with mfgcap
        [prdPlanningAgent addToMfgOrders: outOrder];
   }
   [self->inOrdBuf removeFirst];
 }

 return self;
}

//---get an order from top of inOrdBuf list
-(Order *) getTopOfOrderList: (id) aList {
 Order * anOrder;
 Order * theOrder;
 id index;
  
 index = [aList begin: [self getZone]];
 if ((anOrder = [index next])) {
   theOrder = [anOrder copy: [self getZone]];
   [index drop];
   return theOrder;
 }
 else {
   [index drop];
   return anOrder;
 }
}

//----get the outOrders List
-getOutOrderList {
 return self->outOrders;
}

//---get an order from inOrdBuf --
-getOrder: (Order *) inOrder {
 [self->inOrdBuf addLast: inOrder];
 return self;
}

//---move top of outOrders to the end of sentOrders --
-moveFromOutOrdersToSentOrders {
 Order * anOrder;
 Order * theOrder;
 
 anOrder = [self->outOrders removeFirst];
 theOrder = [anOrder copy: [self getZone]];
 [self->sentOrders addLast: theOrder];

 return self;
}

//------estimate  due date to incoming orders --
-(float) estimateCommitDate: (Order *) anOrder level: (int) upProp {
 float commitDate;
 int demand;
 int receiving;

 demand = anOrder->itemQ;
 
 self->accumuOrder = [invAgent sumOfDemandList: anOrder];
 self->currentInv = [invAgent checkInv: anOrder->itemId];
 self->inOrdering = [self checkInOrdering: anOrder];
 
 if ((self->currentInv - self->accumuOrder) >= demand) { // all from inv
   if (anOrder->wantDate < 0.0)
     commitDate = getCurrentTime();
   else if (anOrder->wantDate >= getCurrentTime())
     commitDate = anOrder->wantDate;
 }
 else { //  from receiving or mfg...
    receiving = demand - (self->currentInv - self->accumuOrder);
    anOrder->itemQ = receiving;
    anOrder->dueDate = -1;
    if ([scnEntity getMfgCap] == 'Y')
      commitDate =[prdPlanningAgent estimateFinishTime: anOrder level: upProp];
    else if ([scnEntity getMfgCap] == 'N')
      commitDate = [self estimateProdArrivalTime: anOrder level: upProp];
 }

 return commitDate;
}

//------estimate the current receiving orders to fulfill order
-(float) estimateProdArrivalTime: (Order *) anOrder level: (int) upProp {
 int amtOutOrders;
 int amtSentOrders;
 float dueDate;
 float deliverTime, availTime;
 float receiving;
 int prodNo;
 int aSupplierNo;
 SCNEntitySwarm * aSupplierE;
 int upShift;

 if ([scnEntity getMfgCap] == 'S') 
   return anOrder->ordArrTime;
    
 receiving = anOrder->itemQ;
 prodNo = anOrder->itemId;

 amtOutOrders = [self checkOrderList: self->outOrders with:  prodNo];
 amtSentOrders = [self checkOrderList: self->sentOrders with: prodNo];
  
 if (receiving <= amtSentOrders)
   dueDate = [self getLatestOrderDueDate: self->sentOrders for: prodNo with: receiving];
 else if (receiving < (amtOutOrders + amtSentOrders)) {
   receiving -= amtSentOrders;
   dueDate = [self getLatestOrderDueDate: self->outOrders for: prodNo with: receiving];
 }
 else { 
   // need to order from upstream suppliers
   if ([scnEntity getMultiSuppliers] == No) {  //MonoSupplier
     aSupplierNo = [scnEntity chooseSupplier: prodNo];
     aSupplierE = [ofpModel allocateEntity: aSupplierNo];
     deliverTime = [aSupplierE getDeliverTime: [scnEntity getScnId]];
     if (upProp == 0) {
       availTime = [scnEntity getDelayTime: aSupplierNo];
       availTime += anOrder->ordArrTime;
     }
     else {
       anOrder->ordArrTime += [scnEntity getDelayTime: aSupplierNo];
       upShift = upProp - 1;
       availTime = [aSupplierE acceptCommitDateQueryMessage: anOrder level: upShift];
     }
     dueDate = deliverTime + availTime;
   }
   else if ([scnEntity getMultiSuppliers] == Yes) {  // MultiSupplier 
     dueDate = [mrpAgent determineCommitDateFromMultiSuppliers: anOrder level: upProp];
   }
 }

 return dueDate;
}


-(float) getLatestOrderDueDate: (id) ordList for: (int) prodId with: (int) receiving {
 
 id index;
 Order * anOrder;
 float procT = 0.0;
 
 index = [ordList begin: [self getZone]];
 while ((anOrder = [index next])) {
   if (anOrder->itemId == prodId) {
     receiving -= anOrder->itemQ;
     if (receiving <= 0)
       procT = anOrder->dueDate;
   }
 }
 [index drop];

 return procT;
}

//------check the inventory availability for fulfilling the order --
-(int) checkAvailInv: (Order *) anOrder {
 int needToOrder;
 
 [traceAgent pHeader: [scnEntity getTC] value:"incoming order =" fmt:Space];
 [traceAgent pOrder: [scnEntity getTC] value:anOrder];
 
 self->accumuOrder = [invAgent sumOfDemandList: anOrder];
 [traceAgent pHeader: [scnEntity getTC] value:"accumuOrder =" fmt:Space];
 [traceAgent pInt: [scnEntity getTC] value:self->accumuOrder fmt: NL];

 self->currentInv = [invAgent checkInv: anOrder->itemId];
 [traceAgent pHeader: [scnEntity getTC] value:"currentInv  =" fmt:Space];
 [traceAgent pInt: [scnEntity getTC] value:self->currentInv fmt: NL];

 self->inOrdering = [self checkInOrdering: anOrder];
 [traceAgent pHeader: [scnEntity getTC] value:"inOrdering =" fmt:Space];
 [traceAgent pInt: [scnEntity getTC] value:self->inOrdering fmt: NL];

 if ((self->currentInv + self->inOrdering - self->accumuOrder) >= anOrder->itemQ)
   needToOrder = 0;
 else
   needToOrder = anOrder->itemQ - (self->currentInv + self->inOrdering - self->accumuOrder);

  return needToOrder;
}

//--------check the number of production in ordering ---
-(int) checkInOrdering: (Order *) anOrder {
 int total = 0;
 
 // check outOrders List
 total = [self checkOrderList: self->outOrders with:  anOrder->itemId];

 // check sentOrders list
 total += [self checkOrderList: self->sentOrders with: anOrder->itemId];

 // check  productionPlan and mfgOrders if withMfgCap= 'Y'
 if ([scnEntity getMfgCap] == 'Y') {
   total += [prdPlanningAgent getNumProdInPlanning: anOrder];
   total += [prdPlanningAgent getNumMfgOrders: anOrder];
 }
   
 return total;
}

//------Quantity of ordList with anOrder
-(int) checkOrderList: (id) ordList with: (int) prodId {
 int amt = 0;
 id index;
 Order * anOrder;
 
 index= [ordList begin: [self getZone]];
 while ((anOrder = [index next])) {
   if (anOrder->itemId == prodId)
    amt += anOrder->itemQ;
 }
 [index drop];

 return amt;
}

//--------check the number of production in ordering ---
-(int) checkInOrderingForMrp: (IdNo *) aComp {
 int total = 0;
  
 // check outOrders List
 total = [self checkOrderList: self->outOrders with: aComp->no];
   
 // check sentOrders List
 total += [self checkOrderList: self->sentOrders with: aComp->no];
   
 return total;
}

//---reduce the sentOrders List while the ordered parts add into stock
-reduceSentOrders: (Part *) aPart {
 id index;
 Order * anOrder;
 int Qleft = aPart->partQ;

 index = [self->sentOrders begin:[self getZone]];
 while ((anOrder = [index next]) && (Qleft > 0)) {
   if (anOrder->itemId == aPart->partId) {
     if (Qleft >= anOrder->itemQ) {
       Qleft -= anOrder->itemQ;
       [index remove];
     }
     else {
       anOrder->itemQ -= Qleft;
       Qleft = 0;
     }
   }
 }
 [index drop];

 return self;
}


//---for reOrderAgent to use --
-moveBackOrdersToOutOrders {
 id index;
 Order * theOrder, * anOrder;

 index = [self->backOrders begin: [self getZone]];
 while ((anOrder = [index next])) {
   theOrder = [anOrder copy: [self getZone]];
   [self->outOrders addLast: theOrder];
 }
 [index drop];
 return self;
}

//---
-moveBackOrdersTomfgOrders {
 id index;
 Order * theOrder, *anOrder;

 index = [self->backOrders begin: [self getZone]];
 while ((anOrder = [index next])) {
   theOrder = [anOrder copy: [self getZone]];
   [prdPlanningAgent addToMfgOrders: theOrder];
 }
 [index drop];
 return self;
}

//---
-removeBackOrders {
 id index;
 Order * anOrder;

 index = [self->backOrders begin: [self getZone]];
 while ((anOrder = [index next])) {
   [index remove];
 }
 [index drop];
 return self;
}

//-- add backOrders to MTS process
-(int) accumuBackOrders: (int) prodNo {
 int amt = 0;
 id index;
 Order * anOrder;

 index = [self->backOrders begin: [self getZone]];
 while ((anOrder = [index next])) {
   if (anOrder->itemId == prodNo)
     amt += anOrder->itemQ;
 }
 [index drop];

 return amt;
}


//----add one order to self->outOrders from Forecasting
-addOneToOutOrders: (Order *) anOrder {
 [self->outOrders addLast: anOrder];
 return self;
}

//----------set the agent which hold inventory for order fulfillment---
-setInvAgent: (id) agent {
 invAgent = agent;
 return self;
}

//----------set the agent which schedules production------------
-setPrdPlanningAgent: (id) agent {
 prdPlanningAgent = agent;
 return self;
}

-setScnEntity: (id) agent {
 scnEntity = agent;
 return self;
}

-setMrpAgent: (id) agent {
 mrpAgent = agent;
 return self;
}

-setOfpModel: (id) agent {
 ofpModel = agent;
 return self;
}
-setTraceAgent: (id) agent {
 traceAgent = agent;
 return self;
}

@end



