// Copyright (C) 1996 Fu-Ren Lin
// ProductionPlanning.m

#import "ProductionPlanning.h"

@implementation ProductionPlanning

+create: (id) aZone {
 ProductionPlanning * newPrdP;
 newPrdP = [super create: aZone];
 newPrdP->buildPlan = [List create: aZone];
 newPrdP->dispatchedPlan = [List create: aZone];
 newPrdP->finishedPlan = [List create: aZone];
 newPrdP->mfgOrders = [List create: aZone];
 return newPrdP;
}

//----initialize the counterForSchedule --
-initCounterForSchedule {
 self->counterForSchedule = 1;
 return self;
}

//-----------scheduling production to build builtPlan-------
-process {
 [traceAgent pHeader: [scnEntity getTC] value: "productionP=" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getScnId] fmt:NL];
  
 [traceAgent pHeader: [scnEntity getTC] value: "CoordSMC=" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getCoordSMC] fmt:NL];
 
 if ([scnEntity getCoordSMC] == Yes)
   [self processSynchronized];
 else if ([scnEntity getCoordSMC] == No)
   [self processNone];

 return self;
}

//-- process without any experimental strategy
-processNone {
 float capacity;
 float usedCap;
 Order * mfgOrder;
 int i;
 int span;
 float startTime, finTime;
 float accumuUsedCap;
 int   completeADay;
 span = PlanHorizone * UNIT; 
 if ([self checkTimeToSchedule] == TRUE) {
   for (i=0; i<PlanHorizone; i++) { // planning
     capacity = UNIT;
     accumuUsedCap = 0.0;
     startTime = counterForSchedule * span + capacity * i;
     finTime = startTime + UNIT;
     completeADay = FALSE;
     while ((completeADay == FALSE) && (capacity > 0.0) && ([self->mfgOrders getCount] > 0)) {
       mfgOrder = [self->mfgOrders getFirst];
       usedCap = [self checkUsedCap: mfgOrder->itemId size: mfgOrder->itemQ];
       if ((mfgOrder->dueDate > finTime) && (usedCap < UNIT))
          completeADay = TRUE;
       else {
         if (usedCap <= capacity) {
           mfgOrder = [self removeTopMfgOrders];
           capacity -= usedCap;
           [self addMfgOrderToPlan: mfgOrder start: startTime end: finTime];
           
           [traceAgent pHeader: [scnEntity getTC] value: "mfgOrder" fmt: Space];
           [traceAgent pOrder: [scnEntity getTC] value: mfgOrder]; 
         }
         else  {
           mfgOrder = [self reduceTopMfgOrders: capacity/usedCap];
           if (mfgOrder->itemQ > 0) {
             [self addMfgOrderToPlan: mfgOrder start: startTime end: finTime];
             
             [traceAgent pHeader: [scnEntity getTC] value: "mfgOrder" fmt: Space];
             [traceAgent pOrder: [scnEntity getTC] value: mfgOrder]; 
           }
           capacity = -1.0;
         }
       } //-- end of else in if ((mfgOrder->dueDate>finTime)&&(usedCap < UNIT))
     }
   }
 }
 return self;
}

//SSSS production planning with "synchronized between material and capacity planning"
-processSynchronized {
 Order * mfgOrder;
 id index;
 id <List> capPlanList;
 id <List> revisedCapPlanList;
 float latestAvailTime;
 Plan * earliestPlan;
 
 index = [self->mfgOrders begin: [self getZone]];
 while ((mfgOrder = [index next])) {
   [traceAgent pHeader: [scnEntity getTC] value: "mfgOrder to productionP =" fmt: Space];
   [traceAgent pOrder:[scnEntity getTC] value: mfgOrder];
   
   latestAvailTime = [mrpAgent checkMatAvail: mfgOrder level: [scnEntity getUpPropLayer]];

   capPlanList = [List create: [self getZone]];
   capPlanList = [capPlanningAgent checkCapacity: mfgOrder];
   if ([capPlanList getCount] > 0) {
     earliestPlan = [capPlanList getFirst];
     if (latestAvailTime <= earliestPlan->scheduledStart) {
       [capPlanningAgent allocateCapacity: capPlanList];
       [self moveCapPlanToPrdPlan: capPlanList];        // o.k.
     }
     else {
       revisedCapPlanList = [List create: [self getZone]];
       revisedCapPlanList = [capPlanningAgent reviseCapAllocation: capPlanList earliest: latestAvailTime wrt: Yes];
       if ([revisedCapPlanList getCount] > 0)
         [self moveCapPlanToPrdPlan: revisedCapPlanList];
     }
   }
   [index remove];
 }
 [index drop];

 return self;
}


//SSSS--move capPlanList to buildPlan ---
-moveCapPlanToPrdPlan: (id) capPlanList {
id capIndex, bIndex;
Plan * aCapPlan, *aBuildPlan;
int found;

 capIndex = [capPlanList begin: [self getZone]];
 while ((aCapPlan = [capIndex next])) {
   aCapPlan->mrpMark = NOTMrp;
   bIndex = [self->buildPlan begin: [self getZone]];
   found = FALSE;
   while ((aBuildPlan = [bIndex next]) && (found == FALSE)) {
     if (aBuildPlan->scheduledStart > aCapPlan->scheduledStart) {
       [bIndex addBefore: aCapPlan];
       found = TRUE;
     }
   }
   if (found == FALSE) 
     [self->buildPlan addLast: aCapPlan];

   [bIndex drop];
 }
 [capIndex drop];

 return self;
}


//------estimate the finished prod to fulfill order --
-(float) estimateFinishTime: (Order *) anOrder level: (int) upProp {
 float finishTime;
 Order * theOrder;
 
  theOrder = [anOrder copy: [self getZone]];
  if (upProp == 0)
    finishTime = [self estimateDueDate: theOrder];
  else {
    if ([scnEntity getCoordSMC] == No)
      finishTime = [self estimateFinishTime_withoutSMC: theOrder level: upProp];
    else if ([scnEntity getCoordSMC] == Yes)
      finishTime = [self estimateFinishTime_withSMC: theOrder level: upProp];
  }

  return finishTime;
}

-(float) estimateFinishTime_withoutSMC: (Order *) anOrder level: (int) upProp {
 float dueDate;
 int amtOfMfgOrders;
 int amtOfBuildPlan;
 int amtOfWaitDispatch;
 int receiving;
 float matAvailTime, mfgTime;
 
   receiving = anOrder->itemQ;
  // amtOfWaitDispatch = [mrpAgent checkProcessedPlanNotDispatched: anOrder->itemId];
   amtOfBuildPlan = [self getNumProdInPlanning: anOrder];
   amtOfMfgOrders = [self getNumMfgOrders: anOrder];
   

   [traceAgent pHeader: [scnEntity getTC] value: "amtOfBuildPlan" fmt: Space];
   [traceAgent pFloat: [scnEntity getTC] value: amtOfBuildPlan fmt:Tab];
   [traceAgent pHeader: [scnEntity getTC] value: "amtOfMfgOrders" fmt: Space];
   [traceAgent pFloat: [scnEntity getTC] value: amtOfMfgOrders fmt:NL];

   if (receiving <= amtOfBuildPlan)   // all from buildPlan
     dueDate = [self getLatestBuildPlanFinishedTime: anOrder->itemId with: receiving]; 
   else if (receiving <= (amtOfBuildPlan + amtOfMfgOrders)) {
     receiving  -= amtOfBuildPlan;
     dueDate = [self getLatestMfgOrdersDueDate: anOrder->itemId with: receiving];
   }
   else {
     receiving -= (amtOfBuildPlan + amtOfMfgOrders);
     dueDate = [self getLastMfgOrderDueDate];
     matAvailTime = [mrpAgent checkMatAvail: anOrder level: upProp];
     mfgTime = [scnEntity getMfgProcTime: anOrder->itemId size: anOrder->itemQ];
     if (dueDate > matAvailTime)
        dueDate += mfgTime;
     else
        dueDate = matAvailTime + mfgTime;
    [traceAgent pHeader: [scnEntity getTC] value: "matAvailTime=" fmt: Space];
    [traceAgent pFloat: [scnEntity getTC] value: matAvailTime fmt:NL];
    [traceAgent pHeader: [scnEntity getTC] value: "dueDate=" fmt: Space];
    [traceAgent pFloat: [scnEntity getTC] value: dueDate fmt:NL];
    [traceAgent pHeader: [scnEntity getTC] value: "mfgTime=" fmt: Space];
    [traceAgent pFloat: [scnEntity getTC] value: mfgTime fmt:NL];
    [traceAgent pHeader: [scnEntity getTC] value: "estimateDueTruMfg" fmt: Space];
    [traceAgent pFloat: [scnEntity getTC] value: dueDate fmt:NL];
   }

   return dueDate;
}

-(float) getLastMfgOrderDueDate {
 Order * mfgOrder;
 
 if ([self->mfgOrders getCount] > 0) {
   mfgOrder = [self->mfgOrders getLast];
   return mfgOrder->dueDate;
 }
 else
   return 0.0;
}


-(float) estimateFinishTime_withSMC: (Order *) anOrder level: (int) upProp {
 id <List> capPlanList;
 id <List> revisedCapPlanList;
 float latestAvailTime;
 Plan * lastCapPlan;
 int onPlan;

 [traceAgent pHeader: [scnEntity getTC] value: "anOrder to estimate =" fmt: Space];
 [traceAgent pOrder:[scnEntity getTC] value: anOrder];
   
 latestAvailTime = [mrpAgent checkMatAvail: anOrder level: upProp];
 onPlan = [mrpAgent checkProcessedPlanNotDispatched: anOrder->itemId] + 
          [self getNumProdInPlanning: anOrder] + 
          [self getNumMfgOrders: anOrder];
  
 if (onPlan >= anOrder->itemQ)   // all on the buildPlan
   return latestAvailTime;
 else 
   anOrder->itemQ -= onPlan;

 capPlanList = [List create: [self getZone]];
 capPlanList = [capPlanningAgent checkCapacity: anOrder];
 if ([capPlanList getCount] > 0) {
   lastCapPlan = [capPlanList getLast];
   if (latestAvailTime <= lastCapPlan->scheduledStart) {
      return lastCapPlan->finishedTime;
   }
   else {
     revisedCapPlanList = [List create: [self getZone]];
     revisedCapPlanList = [capPlanningAgent reviseCapAllocation: capPlanList earliest: latestAvailTime wrt: No];
     if ([revisedCapPlanList getCount] > 0) {
       lastCapPlan = [revisedCapPlanList getLast];
       return lastCapPlan->finishedTime;
     }
     else 
       return (endDay*2.0*UNIT);
   }
 }
 else 
   return (endDay*2.0*UNIT);
}


//-----add one mfgOrder to buildPlan
-addMfgOrderToPlan: (Order *) order start: (float) startTime end: (float) finTime {
 Plan * aPlan;
 id index;

 aPlan = [Plan create: [self getZone]];
 aPlan->prodId = order->itemId;
 aPlan->prodQ = order->itemQ;
 aPlan->scheduledStart = startTime;
 aPlan->finishedTime   = finTime;
 aPlan->mrpMark = NOTMrp;
 [buildPlan addLast: aPlan];
 

 index = [self->buildPlan begin: [self getZone]];
 while ((aPlan = [index next])) {
    [traceAgent pHeader: [scnEntity getTC] value: "buildPlan" fmt: Space];
    [traceAgent pPlan: [scnEntity getTC] value: aPlan]; 
 }
 [index drop];

 return self;
}

//-----remove the top of mfgOrders list (been planned) --
-(Order *) removeTopMfgOrders {
 Order * mfgOrder;
 Order * theMfgOrder;
 
 mfgOrder = [self->mfgOrders removeFirst];
 theMfgOrder = [mfgOrder copy: [self getZone]];

 return theMfgOrder;
}

//------partially plan the top of mfgOrders ----
-(Order *) reduceTopMfgOrders: (float) ratio {
 Order * mfgOrder;
 Order * plannedOrder;
 Order * emptyOrder;

 mfgOrder = [self->mfgOrders getFirst];
 plannedOrder = [mfgOrder copy: [self getZone]];
 plannedOrder->itemQ = mfgOrder->itemQ * ratio;
 mfgOrder->itemQ -=plannedOrder->itemQ;
 if (mfgOrder->itemQ == 0)
   emptyOrder = [self removeTopMfgOrders];

 return plannedOrder;
}

//-----------check the time point for scheduling-----
-(int) checkTimeToSchedule {
 int ready;
 int span;

 span = PlanHorizone * UNIT;
 if (((counterForSchedule * span - getCurrentTime()) <= TimeStep) &&
      ((counterForSchedule * span) >=  getCurrentTime())) {
   counterForSchedule++;
   ready = TRUE;
 }
 else
   ready = FALSE;
 return ready;
}

//---------get the top of buildPlan as a workorder -------------
-(Plan *) getTopWorkOrders {
 Plan * workOrder;
 Plan * aPlan;
 id index;

 index = [self->buildPlan begin: [self getZone]];
 if (( aPlan = [index next])) {
   workOrder = [aPlan copy: [self getZone]];
   [index drop];
   return workOrder;
 }
 else {
  [index drop];
  return aPlan;
 }
}

//---remove the top of buildPlan since it has been moved as a workorder 
// and add into the dispatchedPlan list
-removeTopWorkOrders {
 Plan * workOrder;
 id index;
 
 index = [self->buildPlan begin: [self getZone]];
 if ((workOrder = [index next])) {
   workOrder->finishedTime = getCurrentTime(); // denote the dispatched time
   [dispatchedPlan addLast: workOrder];
   [index remove]; //remove the top of buildPlan
 }
 [index drop];

 return self;
}

//-----------check the expected used capacity for an entry of NeedToManuOrder--
-(float) checkUsedCap: (int) prodNo size: (int) prodQ {
 float procT = 0.0;
  
 procT = [scnEntity getMfgProcTime: prodNo size: prodQ];

 return procT;
}

//-- move entries from dispatchedPlan list to finishedPlan list.
-fulfillBuildPlan: (Plan *) workOrder {
 id index;
 int quanLeft = workOrder->prodQ;
 Plan * disPlan;

 index = [dispatchedPlan begin: [self getZone]];
 while ((quanLeft > 0) && (disPlan = [index next])) {
   if (disPlan->prodId == workOrder->prodId) {
     if (disPlan->prodQ <= quanLeft) {
       quanLeft -= disPlan->prodQ;
       disPlan->finishedTime = getCurrentTime();
       [self addToFinishedPlan: [index remove]];
     }
     else {
       workOrder->finishedTime = getCurrentTime();
       [self->finishedPlan addLast: workOrder];
       disPlan->prodQ -= workOrder->prodQ;
     }
   }
 }
 [index drop];

 return self;
}

//-----move the entry from dispatchedPlan to finishedPlan
-addToFinishedPlan: (Plan *) disPlan {
 Plan * finPlan;
 
   finPlan = [disPlan copy: [self getZone]];
   [self->finishedPlan addLast: finPlan];

 return self;
}

//------add an mfgOrder from OrdM agent and assign due date if needed ---
-addToMfgOrders: (Order *) anOrder {
 int found;
 float newDueDate;
 id index;
 Order * mfgOrder;

 found = FALSE;
 // assign due date
 newDueDate = [self estimateDueDate: anOrder];
 if (newDueDate < anOrder->wantDate)
   anOrder->dueDate = anOrder->wantDate;
 else if (newDueDate > anOrder->dueDate)
   anOrder->dueDate = newDueDate;

 // add into mfgOrders list
 index = [mfgOrders begin: [self getZone]];
 found = FALSE;
 while ((mfgOrder = [index next]) && (found == FALSE)) {
   if (mfgOrder->dueDate > anOrder->dueDate) {
     [index addBefore: anOrder];
     found = TRUE;
   }
 }
 [index drop];

 if (found == FALSE) {
   [self->mfgOrders addLast: anOrder];
 }

 index = [mfgOrders begin: [self getZone]];
 while ((mfgOrder = [index next])) {
   [traceAgent pHeader: [scnEntity getTC] value: "mfgOrder" fmt: Space];
   [traceAgent pOrder: [scnEntity getTC] value: mfgOrder];
 }
 
 return self;
}

//----estimate due date for the coming order
-(float) estimateDueDate: (Order *) anOrder {
 float dueDate;
 float addedTime;
 float totalWork;
 float startT;

 totalWork = [scnEntity getMfgProcTime: anOrder->itemId size: anOrder->itemQ];
 addedTime = [scnEntity getDueDateTightness] * totalWork; // denote the due date tightness
 startT = [self getStartTbyCheckingMfgOrders];
 dueDate =  startT + addedTime;

 return dueDate;
}

-(float) getStartTbyCheckingMfgOrders {
float startT;
Plan * workOrder;
Order *mfgOrder;

 if ([self->mfgOrders getCount] > 0) {
   mfgOrder = [self->mfgOrders getLast];
   startT = mfgOrder->dueDate;
 }
 else if ([self->buildPlan getCount] > 0){
   workOrder = [self->buildPlan getLast];
   startT = workOrder->finishedTime;
 }
 else
   startT = counterForSchedule * PlanHorizone * UNIT;
  
 return startT;
}

//----check a plan for MRP --
-(Plan *) checkBuildPlanForMrp {
 id index;
 Plan * aPlan;
 int found = FALSE;
 int offset = -1;

 index = [self->buildPlan begin: [self getZone]];
 while ((aPlan = [index next]) && (found == FALSE)) {
   offset++;
   if (aPlan->mrpMark == NOTMrp) {
     aPlan->mrpMark = Checking;
     found = TRUE;
   }
 }
 [index drop];
 
 if (found == TRUE)
   aPlan = [self->buildPlan atOffset: offset];
 
 return aPlan;
}

//----check a plan for MRP --
-markBuildPlanForMrp: (char) mark {
 id index;
 Plan * aPlan;
 int found = FALSE;

 index = [self->buildPlan begin: [self getZone]];
 while ((aPlan = [index next]) && (found == FALSE)) {
   if (aPlan->mrpMark == Checking) {
     aPlan->mrpMark = mark;
     found = TRUE;
   }
 }
 [index drop];

 return self;
}

/*
//-----sum of the prodQ in dispatchedPlan with the same prodId with aPlan
-(int) checkDispatchedPlanNotFinished: (Plan *) aPlan {
 int sum = 0;
 id index;
 Plan * disPlan;
 Plan * mrpPlan;

 index = [dispatchedPlan begin: [self getZone]];
 while ((disPlan = [index next])) {
   if (disPlan->prodId == aPlan->prodId) 
     sum += disPlan->prodQ;
 }
 [index drop];
 
// consider entries in buildPlan which are mrped  but not dispatched
 sum += [mrpAgent checkProcessedPlanNotDispatched: aPlan];

 return sum;
}
*/

-(int) getNumProdInPlanning: (Order *) anOrder {
 id index;
 int num = 0;
 Plan * aPlan;

 index = [self->buildPlan begin: [self getZone]];
 while ((aPlan = [index next])) {
   if (aPlan->prodId == anOrder->itemId)
     num += aPlan->prodQ;
 }
 [index drop];
 
 return num;
}

-(int) getNumMfgOrders: (Order *) anOrder {
 id index;
 int num = 0;
 Order * mfgOrder;

 index = [self->mfgOrders begin: [self getZone]];
 while ((mfgOrder = [index next])) {
   if (mfgOrder->itemId == anOrder->itemId)
     num += mfgOrder->itemQ;
 }
 [index drop];
 
 return num;
}

-(float) getLatestMfgOrdersDueDate: (int) prodId with: (int)inOrdering {
 id index;
 Order * mfgOrder;
 float procT = 0.0;
 int found = FALSE;
 float latestTime;

 index = [self->mfgOrders begin: [self getZone]];
 while ((mfgOrder = [index next]) && (found == FALSE)) {
   if (mfgOrder->itemId == prodId) {
     inOrdering -= mfgOrder->itemQ;
     latestTime = mfgOrder->dueDate;
     if (inOrdering <= 0) {
       procT = mfgOrder->dueDate;
       found = TRUE;
     }
   }
 }
 [index drop];

 if (found == TRUE)
   return procT;
 else
   return latestTime;
}

-(float) getLatestBuildPlanFinishedTime: (int) prodId with: (int) inOrdering {
 id index;
 Plan * aPlan;
 float procT = 0.0;
 
 index = [self->buildPlan begin: [self getZone]];
 while ((aPlan = [index next])) {
   if (aPlan->prodId == prodId) {
     inOrdering -= aPlan->prodQ;
     if (inOrdering <= 0)
       procT = aPlan->finishedTime;
   }
 }
 [index drop];

 return procT;
}

//-- know the other agents
-setScnEntity: (id) agent {
 scnEntity = agent;
 return self;
}

-setMrpAgent: (id) agent {
 mrpAgent = agent;
 return self;
}

-setCapPlanningAgent: (id) agent {
 capPlanningAgent = agent;
 return self;
}

-setTraceAgent: (id) agent {
 traceAgent = agent;
 return self;
}

@end
