// Copyright (C) 1996 Fu-Ren Lin
// ReOrdering.m

#import "ReOrdering.h"

@implementation ReOrdering

+create: (id) aZone {
 ReOrdering * newReOrd;
 newReOrd = [super create: aZone];
 newReOrd->eoqList = [List create: aZone];
 newReOrd->ropList = [List create: aZone];
 newReOrd->ropLeadTimeList = [List create: aZone];
 newReOrd->compDQs = [List create: aZone];
 return newReOrd;
}

//----initialize the counterForReOrdering --
-initCounterForReOrdering {
 self->counterForReOrdering = 1;
 return self;
}

//-----------reordering inventory process --
-process {
 Part * anEOQ;
 id index;

 [traceAgent pHeader: [scnEntity getTC] value: "reOrdM=" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getScnId] fmt:NL];

// for Make-To-Stock (MTS) only
 if (([scnEntity getDemandPolicy] != MTS) && 
     ([scnEntity getDemandPolicy] != ATO))
   return self;

// for part souce entity: exit
 if ([scnEntity getMfgCap] == 'S')
   return self;

// for other entities: continue

 index = [self->eoqList begin: [self getZone]];
 while ((anEOQ = [index next])) {
   if ([self checkTimeToReOrdering: anEOQ] == TRUE) { 
     [self reOrderingProcess: anEOQ];
   }
 }
 [index drop];
 
 index = [self->eoqList begin: [self getZone]];
 while ((anEOQ = [index next])) {
  [traceAgent pHeader: [scnEntity getTC] value:"anEOQ=" fmt: Space];
  [traceAgent pPart: [scnEntity getTC] value: anEOQ];
 }

 return self;
}

//---- reOrdering process --
-reOrderingProcess: (Part *) anEOQ {
 Order * anOrder;
 float availTime;

 anOrder = [Order create: [self getZone]];
 anOrder->itemId = anEOQ->partId;
 anOrder->itemQ  = anEOQ->partQ;
 anOrder->ordArrTime = getCurrentTime();
 anOrder->wantDate = -1;
 anOrder->dueDate = -1;
 //anOrder->itemQ += [ordAgent accumuBackOrders: anOrder->itemId];
 if ([scnEntity getDemandPolicy] == MTS)
   availTime = [self getROPLeadTime: anOrder endItem: Yes];
 else if ([scnEntity getDemandPolicy] == ATO)
   availTime = [self getROPLeadTime: anOrder endItem: No];

 //revise rop lead time
 if ([scnEntity getCoordIU] == Yes)
   [self reviseROPLeadTime: anOrder->itemId avail: availTime];  

 anOrder->wantDate = getCurrentTime() + availTime;
 anOrder->dueDate  = anOrder->wantDate;
 anOrder->senderId = [scnEntity getScnId];

 if ([scnEntity getDemandPolicy] == ATO) { 
   [ordAgent removeBackOrders];
   [invAgent addToCompDemandList: anOrder];
   [ordAgent addOneToOutOrders: anOrder];
 }
 else if ([scnEntity getDemandPolicy] == MTS) {
   if ([scnEntity getMfgCap] == 'Y') { //with mfgcap
     [ordAgent removeBackOrders];
     [invAgent addToCompDemandList: anOrder];
     [prdPlanningAgent addToMfgOrders: anOrder];
   } 
   else if ([scnEntity getMfgCap] == 'N') {
     [ordAgent removeBackOrders];
     [invAgent addToCompDemandList: anOrder];
     [ordAgent addOneToOutOrders: anOrder];
   }
 }
 [traceAgent pHeader: [scnEntity getTC] value:"reOrder" fmt: Space];
 [traceAgent pOrder: [scnEntity getTC] value: anOrder];

 return self; 
}

//-----------check the time point for reordering --
-(int) checkTimeToReOrdering: (Part *) anEOQ {
 int ready;
 int span;
 int rop;
 Part * aPart;
 id index;
 int found = FALSE;
 Order * anOrder;
 int accumuCompOrder, currentInv, inOrdering, avail;

 span = ROPCheckH * UNIT;
 if (((counterForReOrdering * span) >=  getCurrentTime()) && ((counterForReOrdering * span - getCurrentTime()) <= TimeStep)) {
   counterForReOrdering++;
   ready = TRUE;
 }
 else
   ready = FALSE;
 
 if (ready == FALSE)
   return ready;
 else {           //-----check the availability of prod --
   anOrder = [Order create: [self getZone]];
   anOrder->itemId = anEOQ->partId;
   anOrder->itemQ =anEOQ->partQ;
   anOrder->ordArrTime = getCurrentTime();
   anOrder->wantDate = -1;
   anOrder->dueDate = -1;
   anOrder->senderId = [scnEntity getScnId];
  
   accumuCompOrder = [invAgent sumOfCompDemandList: anOrder];
   currentInv = [invAgent checkInv: anOrder->itemId];
   inOrdering = [ordAgent checkInOrdering: anOrder];
   
   avail = (currentInv + inOrdering) - accumuCompOrder;

   //revise ROP accord. to demand & lead time
   if (([scnEntity getCoordIU] == Yes) || 
       ([scnEntity getCoordID] == Yes)) 
     [self reviseROP: anEOQ->partId]; 

   index = [self->ropList begin: [self getZone]];
   while ((aPart = [index next]) && (found == FALSE)) {
     if (aPart->partId == anEOQ->partId) {
       rop = aPart->partQ;
       found = TRUE;
     }
   }
   [index drop];

   [traceAgent pHeader: [scnEntity getTC] value: "rop =" fmt: Space];
   [traceAgent pInt: [scnEntity getTC] value: rop fmt: NL];
   [traceAgent pHeader: [scnEntity getTC] value: "avail =" fmt: Space];
   [traceAgent pInt: [scnEntity getTC] value: avail fmt: NL];
   
   if ((found == TRUE) && (avail <= rop))
    ready = TRUE;
   else
    ready = FALSE;

   return ready;
 }
}

-reviseROP: (int) prodNo {
 id index, qIndex, eoqIndex;
 Part * aLT;
 Part * anEOQ, * aDQ, * anEntry;
 int found;
 float newROP, dQ;
 float ropLeadTime;
 id <List> invDQs;

//obtain the leadtime
 index = [self->ropLeadTimeList begin: [self getZone]];
 found = FALSE;
 while ((aLT = [index next]) && (found == FALSE)) {
   if (aLT->partId == prodNo) {
     ropLeadTime = aLT->partArrTime;  // store the lead time ...
     found = TRUE;
   }
 }
 [index drop];

 [traceAgent pHeader: [scnEntity getTC] value: "revise-ropLeadTime=" fmt: Space];
 [traceAgent pFloat: [scnEntity getTC] value: ropLeadTime fmt:NL];

 if ([scnEntity getCoordID] == No) {
   invDQs = [scnEntity getInventoryDQs];
   qIndex = [invDQs begin: [self getZone]];
   eoqIndex = [self->eoqList begin: [self getZone]];
   found = FALSE;
   while ((aDQ = [qIndex next]) && (anEOQ = [eoqIndex next]) &&
                                                (found ==FALSE)) { 
     dQ = aDQ->partQ / EOQPeriod;
     found = TRUE;
   }
   [qIndex drop];
   [eoqIndex drop];
 }
 else if ([scnEntity getCoordID] == Yes) {
   dQ = [self reviseDeltaQ: prodNo];
   [traceAgent pHeader: [scnEntity getTC] value: "revise-deltaQ=" fmt: Space];
   [traceAgent pFloat: [scnEntity getTC] value: dQ fmt:NL];
 }
 
 // new ROP ..
 newROP = ropLeadTime/UNIT * dQ;

 // update ropList
  index = [self->ropList begin: [self getZone]];
  found = FALSE;
  while ((anEntry = [index next]) && (found == FALSE)) {
   if (anEntry->partId == prodNo) {
     anEntry->partQ = newROP;
     anEntry->partArrTime = ropLeadTime;
     [traceAgent pHeader: [scnEntity getTC] value: "ropList=" fmt: Space];
     [traceAgent pPart: [scnEntity getTC] value: anEntry];
     found = TRUE;
   }
 }
 [index drop];

 return self;
}

-reviseROPLeadTime: (int) prodNo avail: (float) newLeadTime {
 id index;
 Part * anEntry;
 int found = FALSE;

 index = [self->ropLeadTimeList begin: [self getZone]];
 while ((anEntry = [index next]) && (found == FALSE)) {
   if (anEntry->partId == prodNo) {
      anEntry->partArrTime = (anEntry->partArrTime * anEntry->partQ +
newLeadTime) / (anEntry->partQ + 1);  // revise lead time...
     anEntry->partQ++;   // counter for revision times
     found = TRUE;
   }
 }
 [index drop];
 return self;
}

//--revise dQ for ROP ...(using InfoSharing downstream)..
-(int) reviseDeltaQ: (int) prodNo{
 int dQ, amt;
 float reviewPeriod;
 id index;
 Part *anEOQ;
 int found = FALSE;
 int offset = -1;
 DemandQuery * aQuery;

 // obtain reviewPeriod...
 index = [self->eoqList begin: [self getZone]];
 while ((anEOQ = [index next]) && (found == FALSE)) {
   offset++;
   if (anEOQ->partId == prodNo) {
     reviewPeriod = anEOQ->partArrTime;  // store reviewperiod
     found = TRUE;
   }
 }
 [index drop];
 
 [traceAgent pHeader: [scnEntity getTC] value: "reviewperiod =" fmt: Space];
 [traceAgent pFloat: [scnEntity getTC] value: reviewPeriod fmt:NL];

 // obtain the demandQ over the reviewPeriod...
 aQuery = [DemandQuery create: [self getZone]];
 aQuery->prodId = prodNo;
 aQuery->startTime = getCurrentTime();
 aQuery->endTime = getCurrentTime() + reviewPeriod;
 aQuery->queryLayer = [scnEntity getDownPropLayer];  //assign the layers
 amt = [self queryDemand: aQuery];

 [traceAgent pHeader: [scnEntity getTC] value: "queryDemand =" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: amt fmt:NL];

 // update the eoqList ...
 anEOQ = [self->eoqList atOffset: offset];
 anEOQ->partQ = amt;

 // return dQ..
 dQ = (amt / reviewPeriod) * UNIT;

 return dQ;
}

//***QUERY downstream entities ...
//---downstream entity demand queries --
-(int) queryDemand: (DemandQuery *) aQuery {
 int amt = 0;
 id <List> customerList;
 IdNo * aCustomerId;
 SCNEntitySwarm * aCustomerE;
 float deliverTime;
 DemandQuery * theQuery;
 int endItem;
 id index;

 if (aQuery->queryLayer == 0) {
   if (([scnEntity getDemandPolicy] == MTS) ||
                  ([scnEntity getDemandPolicy] == ATO))
     amt = [self estimateNextReOrdering: aQuery];  // for MTS or ATO
   else if (([scnEntity getDemandPolicy] == BTF) ||
                   ([scnEntity getDemandPolicy] == MTO))
     amt = [forecastingAgent foreSeeDemand: aQuery]; // for BTF or MTO
 }
 else {
   customerList = [scnEntity getCustomerList];
   index = [customerList begin: [self getZone]];
   while ((aCustomerId = [index next])) {
     aCustomerE = [ofpModel allocateEntity: aCustomerId->no];
    
     theQuery = [aQuery copy: [self getZone]];
     deliverTime = [scnEntity getDeliverTime: aCustomerId->no];
     theQuery->startTime += deliverTime;
     theQuery->endTime += deliverTime;
     theQuery->queryLayer--;
     if ([scnEntity getMfgCap] == 'Y') {
       endItem = [scnEntity convertFromCompToEndItem: aQuery->prodId];
       theQuery->prodId = endItem;
     }
     amt += [aCustomerE acceptDownDemandQuery: theQuery];
   }
   [index drop];
 }
  
 return amt;
}

//--- estimate next reordering time and amount --
-(int) estimateNextReOrdering: (DemandQuery *) aQuery {
  int amt = 0;
  id index;
  Part * anEOQ, * theEOQ;
  int found = FALSE;
  int dQ, avail;
  float timeLine;
  int expectedConsumption, accumuCompOrder, currentInv, inOrdering;
  Order * anOrder;

  index = [self->eoqList begin: [self getZone]];
  while ((anEOQ = [index next]) && (found == FALSE)) {
    if (anEOQ->partId == aQuery->prodId) {
       dQ = (anEOQ->partQ/anEOQ->partArrTime) * UNIT;  // DQ/period
       theEOQ = [anEOQ copy: [self getZone]];
       found = TRUE;
    }
  }
  [index drop];
  timeLine = aQuery->endTime - getCurrentTime();
  expectedConsumption = timeLine * dQ;
 
   anOrder = [Order create: [self getZone]];
   anOrder->itemId = theEOQ->partId;
   anOrder->itemQ =theEOQ->partQ;
   anOrder->ordArrTime = getCurrentTime();
   anOrder->wantDate = -1;
   anOrder->dueDate = -1;
   anOrder->senderId = [scnEntity getScnId];
  
  accumuCompOrder = [invAgent sumOfCompDemandList: anOrder];
  currentInv = [invAgent checkInv: anOrder->itemId];
  inOrdering = [ordAgent checkInOrdering: anOrder];
   
  avail = (currentInv + inOrdering) - accumuCompOrder;
  if (avail < expectedConsumption)
    amt = theEOQ->partQ;

 return amt;

}



//-- for components--
-calcurateCompEOQ: (id) invDQs{
 id index1, index2, index3;
 Part * aCompDQ, * theCompDQ;
 Part * aProdDQ;
 id <List> compList;
 IdNo * aComp;
 int found;

 // obtain entries of compDQs
 index1 = [invDQs begin: [self getZone]];
 while ((aProdDQ = [index1 next])) {
   compList = [ofpModel getCompList: aProdDQ->partId];
   index2 = [compList begin: [self getZone]];
   while ((aComp = [index2 next])) {
     index3 = [self->compDQs begin: [self getZone]];
     found = FALSE;
     while ((aCompDQ = [index3 next]) && (found == FALSE)) {
       if (aCompDQ->partId == aComp->no) {
         aCompDQ->partQ += aProdDQ->partQ;  // share the same components
         found = TRUE;
       }
     }
     [index3 drop];
     if (found == FALSE) { 
       theCompDQ = [aProdDQ copy: [self getZone]];
       theCompDQ->partId = aComp->no;
       [self->compDQs addLast: theCompDQ];
     }
   }
   [index2 drop];
  }
  [index1 drop];
  
  // calcurate eoq for components --
  [self calcurateEOQ: self->compDQs endItem: No];
  
 return self;
}

//-----calcureate EOQ (at initial state)
-calcurateEOQ: (id) invDQs endItem: (int) isEndItem {
 Part * anEOQ;
 Part * aDQ;
 float ordCost, invCost, quan;
 id index;
 float freq;

 index = [invDQs begin: [self getZone]];
 while ((aDQ = [index next])) {
   anEOQ = [Part create: [self getZone]];
   anEOQ->partId = aDQ->partId;
   ordCost = [scnEntity getOrderCost: aDQ->partId endItem: isEndItem];
   invCost = [scnEntity getInvCost: aDQ->partId];
   quan = aDQ->partQ;
   anEOQ->partQ = sqrt((2.0 * quan * ordCost / invCost)); //EOQ model
   freq = quan / anEOQ->partQ;
   anEOQ->partArrTime = (EOQPeriod * UNIT)/freq; // store the EOQ period...
   [traceAgent pHeader: [scnEntity getTC] value: "aPart EOQ =" fmt: Space];
   [traceAgent pPart: [scnEntity getTC] value: anEOQ];
   [self->eoqList addLast: anEOQ];
 }
 [index drop];

  return self;
}

// calcurate ROP for components --
-calcurateCompROP {
 [self calcurateROP: self->compDQs endItem: No];
 return self;
}

-getCompDQs {
 return self->compDQs;
}

//-----calcurate ROP = dxL (d=daily demand, L=lead time)
-calcurateROP: (id) invDQs endItem: (int) isEndItem {
int dQ;
float leadTime;
Part * aDQ;
Part * anEOQ;
Part * aROP;
id qIndex, eoqIndex;
Part * leadEntry;
Order * anOrder;

 qIndex = [invDQs begin: [self getZone]];
 eoqIndex = [self->eoqList begin: [self getZone]];
 while ((aDQ = [qIndex next]) && (anEOQ = [eoqIndex next])) { 
   dQ = aDQ->partQ / EOQPeriod;
   
   [traceAgent pHeader: [scnEntity getTC] value: "initialdQ=" fmt: Space];
   [traceAgent pInt: [scnEntity getTC] value: dQ fmt:NL];

   anOrder = [Order create: [self getZone]];
   anOrder->itemId = aDQ->partId;
   anOrder->itemQ  = [self getEOQ: aDQ->partId];
   anOrder->ordArrTime = getCurrentTime();
   leadTime = [self getROPLeadTime: anOrder endItem: isEndItem];

   leadEntry = [aDQ copy: [self getZone]];  // for ropleadtime entries
   leadEntry->partArrTime = leadTime;
   leadEntry->partQ = 1;
   [self->ropLeadTimeList addLast: leadEntry];

   aROP = [Part create: [self getZone]];
   aROP->partId = aDQ->partId;
   aROP->partQ = dQ * leadTime/UNIT;
   aROP->partArrTime = leadTime;
   [traceAgent pHeader: [scnEntity getTC] value: "aPart ROP =" fmt: Space];
   [traceAgent pPart: [scnEntity getTC] value: aROP];
   [self->ropList addLast: aROP];
 }
 [qIndex drop];
 [eoqIndex drop];
 
 return self;
}

// obtain the eoq for a prod.
-(int) getEOQ: (int) prodNo {
 int eoq;
 id index;
 Part * anEOQ;
 int found = FALSE;

 index = [self->eoqList begin: [self getZone]];
 while ((anEOQ = [index next]) && (found == FALSE)) {
   if (anEOQ->partId == prodNo) {
     eoq = anEOQ->partQ;
     found = TRUE;
   }
 }
 [index drop];

 return eoq;
}

//-----obtain the lead time for a product
-(float) getROPLeadTime: (Order *) anOrder endItem: (int) isEndItem {
float leadTime, availTime;

 if (([scnEntity getMfgCap] == 'Y') && (isEndItem == Yes))
   availTime = [prdPlanningAgent estimateFinishTime: anOrder level: [scnEntity getUpPropLayer]];
 else 
   availTime = [ordAgent estimateProdArrivalTime: anOrder level: [scnEntity getUpPropLayer]];

 leadTime = availTime - getCurrentTime();

 [traceAgent pHeader: [scnEntity getTC] value: " getROPLeadTime=" fmt: Space];
 [traceAgent pFloat: [scnEntity getTC] value: leadTime fmt:NL];
 
 return leadTime;
}

//-----get eoqList
-getEoqList {
 id <List> aEoqList;

 aEoqList = [self->eoqList copy: [self getZone]];
 return aEoqList;
}


//-----set up agents
-setInvAgent: (id) agent {
 invAgent = agent;
 return self;
}

-setOrdAgent: (id) agent {
 ordAgent = agent;
 return self;
}

-setMrpAgent: (id) agent {
 mrpAgent = agent;
 return self;
}

-setPrdPlanningAgent: (id) agent {
 prdPlanningAgent = agent;
 return self;
}

-setScnEntity: (id) agent {
 scnEntity = agent;
 return self;
}

-setOfpModel: (id) agent {
 ofpModel = agent;
 return self;
}

-setForecastingAgent: (id) agent {
 forecastingAgent = agent;
 return self;
}

-setTraceAgent: (id) agent {
 traceAgent = agent;
 return self;
}

@end

