// Copyright (C) 1996 Fu-Ren Lin
// SCNEntity.h
// SCNEntity describes the entity properties and schedule for a SCN entity

#import "global.h"
#import "OrderManagement.h"
#import "InventoryManagement.h"
#import "Factory.h"
#import "Dispatching.h"
#import "MaterialPlanning.h"
#import "ProductionPlanning.h"
#import "OFPModelSwarm.h"
#import "Forecasting.h"
#import "ReOrdering.h"
#import "ScnManagement.h"
#import "CapacityPlanning.h"

@interface SCNEntitySwarm: Swarm {
 // properties of the entity
 int  scnId;
 char withMfgCap;
 int multiSuppliers;
 float TWK;      //due date tightness factor
 
 // lists for entity properties
 id <List> suppliers;
 id <List> customers;
 id <List> agentList;
 id <List> orderDelayTs;
 id <List> partDeliverTs;
 id <List> prodIds;
 id <List> mfgProcTs;
 id <List> orderCosts;
 id <List> inventoryCosts;
 id <List> inventoryDQs;

 // lists to maintain information and material flow
 id <List> inOrders;
 id <List> inParts;

 // upper layer agent
 id ofpModel;

 // internal agents 
 id ordM, invM, scnM, reOrdM, forecast, factory, dispatch, productionP, materialP, capacityP;

 // SCN's actions and schedule
 id scnEActions;
 id scnESchedule;

 //scnE's demand management policy
 int demandPolicy;
 
 // scnE's strategies
 int coordIU;
 int coordID;
 int coordSMC;
 int coordDBC;
 int upPropLayer;
 int downPropLayer;

 id traceAgent;
 int tC;
}

+create: (id) aZone;
//---related messages for building objects
-buildObjects;
-assignDemandPolicy: (int) policy;
-(int) getDemandPolicy;

-assignCoordIU: (int) strategy;
-assignCoordID: (int) strategy;
-assignCoordSMC: (int) strategy;
-assignCoordDBC: (int) strategy;
-assignUpPropLayer: (int) strategy;
-assignDownPropLayer: (int) strategy;

-(int) getCoordIU;
-(int) getCoordID;
-(int) getCoordSMC;
-(int) getCoordDBC;
-(int) getUpPropLayer;
-(int) getDownPropLayer;

-denoteAgentRelation;
-setScnId: (int) anId;
-(int) getScnId;
-setMfgCap: (char) capacity;
-(char) getMfgCap;
-assignDueDateTightness: (float) k;
-(float) getDueDateTightness;
-assignMultiSuppliers: (int) YesNo;
-(int) getMultiSuppliers;
-setTraceCode: (int) code;
-(int) getTC;
-setOfpModel: (id) aSwarm;
-addOneSupplier: (int) anId;
-addOneCustomer: (int) anId;
-addOneOrderDelayT: (float) aTime;
-addOnePartDeliverT: (float) aTime;
-addOneProdId: (int) anId;
-addOneMfgProcT: (float) aTime;
-addOneOrderCost: (float) aCost at: (int) offset;
-addOneInvCost: (float) aCost types: (id) typeList at: (int) offset;
-addOneProdDQ: (int) aDQ at: (int) offset;

//---related messages for building actions
-buildActions;

// recognize an order from the orders list
-receiveOrder;

// recognize an order from the orders list
-receivePart;

// send an order to upperstream entities
-orderProduct;
-orderFromSingleSupplier: (Order *) anOrder;
-(int) chooseSupplier: (int) aProdNo;
-chooseMultiSuppliers: (int) aProdNo;  
-(float) getDelayTime: (int) aSupplierNo;
-(float) getDelayTime: (int) senderNo to: (int) receiverNo;

// ship out the finished products
-shipPart;
-getProdIds; // for MfgCap = 'Y'
-getInvTypes; // for MfgCap = 'N'
-addOnePartToInParts: (Part *) newPart;
-addOneOrderToInOrders: (Order *) newOrder;
-(float) getDeliverTime: (int) aCustomerNo;
-(float) getOrderDelayAndPartDeliverWithSupplier: (int) compNo;
-(float) getMfgProcTime: (int) prodId size: (int) prodQ;

-activateIn: (id) swarmContext;

// methods for statistics
-statProcess;
-(float) getCapUtilization;

// method for debugging
-printSuppliers;
-printCustomers;

// methods for forcasting
-(int) belongToProdGroup: (int) prodId;
-getProdGroup;

// methods for reordering
-(float) getOrderCost: (int) prodId endItem: (int) isEndItem;
-(float) getOrderCostWithoutMfgCap: (int) prodId;
-(float) getOrderCostWithMfgCap: (int) prodId;
-(float) getInvCost: (int) prodId;
-getInventoryDQs;

// methods for scnManagement
-getSupplierList;
-receiveQuery: (Query *) aQuery;
-receiveAward: (Query *) anAward;
-receiveBid: (Bid *) aBid;


//methods for mrp for inforSharUp strategy
-(float) acceptCommitDateQueryMessage: (Order *) anOrder level: (int) upProp;

//methods for downstream infoshare
-getCustomerList;
-(int) acceptDownDemandQuery: (DemandQuery *) aQuery;
-(int) convertFromCompToEndItem: (int) aCompNo;
-getInventoryDQs;
-acceptComputeActualDemandQuery: (int) downProp from: (float) stime to: (float) etime prodSet: (id) prodList;
@end




