//Copyright (C) 1996 Fu-ren Lin
//SCNEntitySwarm.m

#import "SCNEntitySwarm.h"

@implementation SCNEntitySwarm
+create: aZone {
  SCNEntitySwarm * newScnE;
  newScnE = [super create: aZone];
  newScnE->suppliers    = [List create: aZone];
  newScnE->customers    = [List create: aZone];
  newScnE->agentList    = [List create: aZone];
  newScnE->orderDelayTs = [List create: aZone];
  newScnE->partDeliverTs= [List create: aZone];
  newScnE->prodIds      = [List create: aZone];
  newScnE->mfgProcTs    = [List create: aZone];
  newScnE->orderCosts   = [List create: aZone];
  newScnE->inventoryCosts   = [List create: aZone];
  newScnE->inventoryDQs  = [List create: aZone];
 
  //
  newScnE->inOrders     = [List create: aZone];
  newScnE->inParts      = [List create: aZone];
  
  return newScnE;
}

-buildObjects {
 [super buildObjects];

 self->traceAgent = [Trace create: [self getZone]];
 self->capacityP = [CapacityPlanning create: [self getZone]];
 [self->capacityP initialCap];
 
 self->ordM = [OrderManagement create: [self getZone]];
 [self->agentList addLast: self->ordM];

 self->invM = [InventoryManagement create: [self getZone]];
 [self->agentList addLast: self->invM];

 self->forecast = [Forecasting create: [self getZone]];
 [self->forecast initCounterForForecasting];
 [self->agentList addLast: self->forecast];

 self->scnM = [ScnManagement create: [self getZone]];
 [self->scnM initBiddingCounter];
 [self->agentList addLast: self->scnM];
 
 self->reOrdM = [ReOrdering create: [self getZone]];
 [self->reOrdM initCounterForReOrdering];
 [self->agentList addLast: self->reOrdM];

 if (self->withMfgCap == 'Y') {
   self->dispatch    = [Dispatching create: [self getZone]];
   [self->agentList addLast: self->dispatch];

   self->factory = [Factory create: [self getZone]];
   [self->factory setInitState];
   [self->agentList addLast: self->factory];

   self->productionP = [ProductionPlanning create: [self getZone]];
   [self->productionP initCounterForSchedule];
   [self->agentList addLast: self->productionP];

   self->materialP   = [MaterialPlanning create: [self getZone]];
   [self->agentList addLast: self->materialP];
 } 
 
 [self denoteAgentRelation]; //help to identify other agents

 if (self->demandPolicy == MTS) {
    // calcurate EOQ, ROP for ReOrdering agent
    [self->reOrdM calcurateEOQ: self->inventoryDQs endItem: Yes];
    [self->reOrdM calcurateROP: self->inventoryDQs endItem: Yes];
    // assign the initial stock size
    [self->invM initInStock];
 }
 else if (self->demandPolicy == ATO) {
  if (self->withMfgCap == 'N') {
    // calcurate EOQ, ROP for ReOrdering agent
    [self->reOrdM calcurateEOQ: self->inventoryDQs endItem: Yes];
    [self->reOrdM calcurateROP: self->inventoryDQs endItem: Yes];
    // assign the initial stock size
    [self->invM initInStock];
  }
  else if (self->withMfgCap == 'Y') {
   // calcurate component EOQ, ROP for ReOrdering agent
   [self->reOrdM calcurateCompEOQ: self->inventoryDQs];
   [self->reOrdM calcurateCompROP];
   // assign the initial stock size
   //[self->invM initInStock];
  }
 }
 return self;
}

-buildActions {

  [super buildActions];
  scnEActions = [ActionGroup create: [self getZone]];

  [scnEActions createActionTo: self message: M(receiveOrder)];
  [scnEActions createActionTo: self message: M(receivePart)];
  [scnEActions createActionTo: self message: M(orderProduct)];
  [scnEActions createActionTo: self message: M(shipPart)];
  [scnEActions createActionForEach: agentList message: M(process)];

  scnESchedule = [Schedule createBegin: [self getZone]];
  [scnESchedule setRepeatInterval: TimeStep];
  scnESchedule = [scnESchedule createEnd];
  [scnESchedule at: 0 createAction: scnEActions];

  return self;
}

//-----------------
-activateIn: (id) swarmContext {
  [super activateIn: swarmContext];
  
  [scnESchedule activateIn: self];

  return [self getSwarmActivity];
}

//xxxxxx--------scnEntity level messages --
//---get an order from the orders list---
-receiveOrder {
 Order * order;
 id index;

 index = [self->inOrders begin: [self getZone]];
 while ((order = [index next])) {
   if (order->ordArrTime <= getCurrentTime()) {
     [self->ordM getOrder: order];
     [index remove];
     [traceAgent pHeader: tC value: "receive ORDER: scnE" fmt: Space];
     [traceAgent pInt:tC value: self->scnId fmt: Tab];
     [traceAgent pOrder:tC value: order];
   }
  }
  return self;
}

//---get a part from the parts list ---
-receivePart {
 Part * part;
 id index;

 index = [self->inParts begin: [self getZone]];
 while ((part = [index next])) {
   if (part->partArrTime <= getCurrentTime()) {
     [self->invM getPart: part];
     [index remove];
     [traceAgent pHeader: tC value: "receivePART= scnE"  fmt: Space];
     [traceAgent pInt:tC value: self->scnId fmt: Tab];
     [traceAgent pPart:tC value: part];
   }
 }

 return self;
}

//---send an order to upperstream entities----
-orderProduct {
Order * anOrder;

// for the part sourcing entity
 if (self->withMfgCap == 'S') 
   return self;

// for the normal entities

 while ((anOrder = [self->ordM getTopOfOrderList: [self->ordM getOutOrderList]])){
   if ((self->coordDBC == Yes) && (self->multiSuppliers == Yes))
     [self->scnM acceptDemandOrder: anOrder];
   else
     [self orderFromSingleSupplier: anOrder];
   
   [self->ordM moveFromOutOrdersToSentOrders];
 }
 return self;
}

//---order products from single supplier
-orderFromSingleSupplier: (Order *) anOrder {
 int num;
 SCNEntitySwarm * supplierE;
 IdNo * supplierId;
 int supplierNo;
 Ttime * delayT;
 float delayTime;
 float deliverTime;

 num = [self->suppliers getCount];
 if (num == 1) {
   supplierId = [self->suppliers getFirst];
   supplierNo = supplierId->no;
   delayT = [self->orderDelayTs getFirst];
   delayTime = delayT->ttime;
 }
 else if (num > 1) {
   supplierNo = [self chooseSupplier: anOrder->itemId];
   delayTime = [self getDelayTime: supplierNo];
 }
 anOrder->ordArrTime += delayTime;
 anOrder->senderId = self->scnId;
   
 supplierE = [ofpModel allocateEntity: supplierNo];
 deliverTime = [supplierE getDeliverTime: self->scnId];
 anOrder->wantDate -= deliverTime;
 anOrder->dueDate -= deliverTime;

 [traceAgent pHeader: tC value: "scnE"  fmt: Space];
 [traceAgent pInt:tC value: self->scnId fmt: Space];
 [traceAgent pHeader: tC value: "sendORDERto scnE"  fmt: Space];
 [traceAgent pInt:tC value: [supplierE getScnId] fmt: Tab];
 [traceAgent pOrder:tC value: anOrder];

 [supplierE addOneOrderToInOrders: anOrder];

 return self;
}

//-------add an order into a supplierE --
-addOneOrderToInOrders: (Order *) newOrder {
 id index;
 Order * anOrder;
 int found = FALSE;

 index = [self->inOrders begin: [self getZone]];
 while ((anOrder = [index next]) && (found == FALSE)) {
   if (anOrder->ordArrTime > newOrder->ordArrTime) {
     [index addBefore: newOrder];
     found = TRUE;
   }
 }
 [index drop];

 if (found == FALSE) {
   [self->inOrders addLast: newOrder];
 }

 return self;
}

//MonoSupplier--select a supplier to deliver a ProdNo (single supplier case)
-(int) chooseSupplier: (int) aProdNo {
IdNo * aSupplierId;
int theSupplierNo = -1;
SCNEntitySwarm * aSupplierE;
IdNo * aProdId;
id index1, index2;
int found = FALSE;
id <List> prodList;
  
  
  index1 = [self->suppliers begin: [self getZone]];
  while ((aSupplierId = [index1 next]) && (found == FALSE)) { 
    aSupplierE = [ofpModel allocateEntity: aSupplierId->no];
    
    if ([aSupplierE getMfgCap] == 'Y')
      prodList = [aSupplierE getProdIds];
    else if (([aSupplierE getMfgCap] == 'N')||([aSupplierE getMfgCap] == 'S'))
      prodList = [aSupplierE getInvTypes];

    index2 = [prodList begin: [self getZone]];
    while ((aProdId = [index2 next])) {
      if (aProdId->no == aProdNo) {
        found = TRUE;
        theSupplierNo = aSupplierId->no;
      }
    }
    [index2 drop];
  }
  [index1 drop];

  return theSupplierNo;
}

//MultiSuppliers--select suppliers (Multiple suppliers case)--
-chooseMultiSuppliers: (int) aProdNo {
id <List> supplierList;
IdNo * aSupplierId, * theSupplierId;
SCNEntitySwarm * aSupplierE;
IdNo * aProdId;
id index1, index2;
id <List> prodList;
int found;
  
  supplierList = [List create: [self getZone]];
  index1 = [self->suppliers begin: [self getZone]];
  while ((aSupplierId = [index1 next])) { 
    aSupplierE = [ofpModel allocateEntity: aSupplierId->no];
    
    if ([aSupplierE getMfgCap] == 'Y')
      prodList = [aSupplierE getProdIds];
    else if (([aSupplierE getMfgCap] == 'N')||([aSupplierE getMfgCap] == 'S'))
      prodList = [aSupplierE getInvTypes];

    index2 = [prodList begin: [self getZone]];
    found = FALSE;
    while ((aProdId = [index2 next]) && (found == FALSE)) {
      if (aProdId->no == aProdNo) {
        found = TRUE;
        theSupplierId = [IdNo create: [self getZone]];
        theSupplierId->no = aSupplierId->no;
        [supplierList addLast: theSupplierId];
      }
    }
    [index2 drop];
  }
  [index1 drop];

  return supplierList;
}

//---the information transfer between two entity (from downstream to upstream)
-(float) getDelayTime: (int) senderNo to: (int) receiverNo {
 SCNEntitySwarm * senderE;
 float delayTime;

 if ([self getMfgCap] == 'S')
   return Infinite;

 senderE = [SCNEntitySwarm create: [self getZone]];
 senderE = [ofpModel allocateEntity: senderNo];
 delayTime = [senderE getDelayTime: receiverNo];
 return delayTime;
}

//-----get order delay time
-(float) getDelayTime: (int) aSupplierNo {
IdNo * supplierId;
id index;
int found = FALSE;
int offset = -1;
Ttime * delayT;

  index = [self->suppliers begin: [self getZone]];
  while ((supplierId = [index next]) && (found == FALSE)) { 
    offset++;
    if (supplierId->no == aSupplierNo)
      found = TRUE;
  }
  [index drop];
  delayT = [self->orderDelayTs atOffset: offset];
  return delayT->ttime;
}

//-------get the prodIds list if withMfgCap == 'Y'
-getProdIds {
 id <List> prodList;

 prodList = [self->prodIds copy: [self getZone]];

/* 
 index = [prodList begin: [self getZone]];
 while (( aProd = [index next]))
   printf("aProd = %d \t", aProd->no);
*/

 return prodList;
}

//-------get the invTypes list if withMfgCap == 'N'
-getInvTypes {
 id index;
 IdNo * prodId;
 InvCost * aCost;
 id <List> prodList;

 prodList = [List create: [self getZone]]; 
 index = [self->inventoryCosts begin:[self getZone]];
 while ((aCost = [index next])) {
   prodId = [IdNo create: [self getZone]];
   prodId->no = aCost->prodId;
   [prodList addLast: prodId];
 }
 [index drop];

 return prodList;
}

//---ship out the finished products----
-shipPart {
 SCNEntitySwarm * customerE;
 Part * aPart;
 float deliverTime;
 IdNo * customerId; 

 while ((aPart = [self->invM getTopOfPartList: [self->invM getOutPartList]])) {
   customerId = [self->invM  getTopOfReceivers];
   if (customerId->no < 0) {
     [self->invM moveFromOutPartsToSentParts];
   }
   else if ((deliverTime = [self getDeliverTime: customerId->no])) {
     aPart->partArrTime += deliverTime;
     customerE = [ofpModel allocateEntity: customerId->no];
     
     [traceAgent pHeader: tC value: "scnE"  fmt: Space];
     [traceAgent pInt:tC value: self->scnId fmt: Space];
     [traceAgent pHeader: tC value: "sendPARTto scnE"  fmt: Space];
     [traceAgent pInt:tC value: [customerE getScnId] fmt: Tab];
     [traceAgent pPart:tC value: aPart];

     [customerE addOnePartToInParts: aPart];
     [self->invM moveFromOutPartsToSentParts];
   }
 }

 return self;
}

//-------add a part into a customerE --
-addOnePartToInParts: (Part *) newPart {
 id index;
 Part * aPart;
 int found = FALSE;

 index = [self->inParts begin: [self getZone]];
 while ((aPart = [index next]) && (found == FALSE)) {
   if (aPart->partArrTime > newPart->partArrTime) {
     [index addBefore: newPart];
     found = TRUE;
   }
 }
 if (found == FALSE) {
   [self->inParts addLast: newPart];
 }
 [index drop];

 return self;
}


//---get the part deliver time to the customer
-(float) getDeliverTime: (int) aCustomerNo {
 Ttime * deliverT;
 id index;
 IdNo * aCustomer;
 int found = FALSE;
 int offset = -1;

   index = [self->customers begin: [self getZone]];
   while ((aCustomer = [index next]) && (found == FALSE)) {
     offset++;
     if (aCustomer->no == aCustomerNo) 
       found = TRUE;
   }
   [index drop];

   if (found == TRUE) {
     deliverT = [self->partDeliverTs atOffset: offset];
     return deliverT->ttime;
   }
   else
     return -1.0;
}

//---- get deliver time for a part at aComp --
-(float) getOrderDelayAndPartDeliverWithSupplier: (int) compNo {
  int theSupplierNo;
  SCNEntitySwarm * theSupplierE;
  float deliverTime;
  float delayTime;

 // choose the supplier
 if ([self getMfgCap] == 'S')
   return Infinite;

 theSupplierNo = [self chooseSupplier: compNo];
 theSupplierE = [ofpModel allocateEntity: theSupplierNo];

 // get the part deliver time
   deliverTime = [theSupplierE getDeliverTime: self->scnId];
  
 // get the order delay time
   delayTime = [self getDelayTime: theSupplierNo];

  return (deliverTime + delayTime);

} 
   
//-----get the mfg process time---
-(float) getMfgProcTime: (int) prodId size: (int) prodQ {
 float procTime;
 int found = FALSE;
 IdNo * pId;
 Ttime * pTime;
 id idIndex, procIndex;
 
 idIndex = [self->prodIds begin: [self getZone]];
 procIndex = [self->mfgProcTs begin: [self getZone]];
 while ((pId = [idIndex next]) && (pTime = [procIndex next]) && (found == FALSE)) {
   if (pId->no == prodId) {
     procTime = pTime->ttime * prodQ;
     found = TRUE;
   }
 } 
 [idIndex drop];
 [procIndex drop];
 
 return procTime;
}


//aaaaa---- construct the acquaintance list for each agent
-denoteAgentRelation {
 // forecasting
 [self->forecast setOrdAgent: self->ordM];
 [self->forecast setInvAgent: self->invM];
 [self->forecast setScnEntity: self];
 [self->forecast setOfpModel: self->ofpModel];
 [self->forecast setPrdPlanningAgent: self->productionP]; 
 [self->forecast setTraceAgent: self->traceAgent];

 // reOrdM
 [self->reOrdM setOrdAgent: self->ordM];
 [self->reOrdM setInvAgent: self->invM];
 [self->reOrdM setMrpAgent: self->materialP]; 
 [self->reOrdM setPrdPlanningAgent: self->productionP]; 
 [self->reOrdM setOfpModel: self->ofpModel];
 [self->reOrdM setScnEntity: self];
 [self->reOrdM setTraceAgent: self->traceAgent];
 [self->reOrdM setForecastingAgent: self->forecast];

 // scnM
 [self->scnM setOrdAgent: self->ordM];
 [self->scnM setInvAgent: self->invM];
 [self->scnM setMrpAgent: self->materialP]; 
 [self->scnM setPrdPlanningAgent: self->productionP]; 
 [self->scnM setOfpModel: self->ofpModel];
 [self->scnM setScnEntity: self];
 [self->scnM setTraceAgent: self->traceAgent];

 // ordM
 [self->ordM setInvAgent: self->invM];
 [self->ordM setScnEntity: self];
 [self->ordM setOfpModel: self->ofpModel];
 [self->ordM setPrdPlanningAgent: self->productionP]; 
 [self->ordM setMrpAgent: self->materialP]; 
 [self->ordM setTraceAgent: self->traceAgent];

 // invM
 [self->invM setOrdAgent: self->ordM];
 [self->invM setScnEntity: self];
 [self->invM setOfpModel: self->ofpModel];
 [self->invM setReOrdAgent: self->reOrdM];
 [self->invM setTraceAgent: self->traceAgent];
 
 // factory
 [self->factory setInvAgent: self->invM];
 [self->factory setPrdPlanningAgent: self->productionP];
 [self->factory setScnEntity: self];
 [self->factory setOfpModel: self->ofpModel];
 [self->factory setTraceAgent: self->traceAgent];

 //capacityP
 [self->capacityP setPrdPlanningAgent: self->productionP];
 [self->capacityP setScnEntity: self];
 [self->capacityP setTraceAgent: self->traceAgent];

 // dispatch	
 [self->dispatch setFactoryAgent: self->factory];
 [self->dispatch setInvAgent: self->invM];
 [self->dispatch setScnEntity: self];
 [self->dispatch setPrdPlanningAgent: self->productionP];
 [self->dispatch setMrpAgent: self->materialP];
 [self->dispatch setTraceAgent: self->traceAgent];
 
 // productionP
 [self->productionP setScnEntity: self];
 [self->productionP setMrpAgent: self->materialP];
 [self->productionP setTraceAgent: self->traceAgent];
 [self->productionP setCapPlanningAgent: self->capacityP];

 // materialP
 [self->materialP setOrdAgent: self->ordM];
 [self->materialP setInvAgent: self->invM];
 [self->materialP setPrdPlanningAgent: self->productionP];
 [self->materialP setScnEntity: self];
 [self->materialP setOfpModel: self->ofpModel];
 [self->materialP setTraceAgent: self->traceAgent];
 //[self->materialP setCapPlanningAgent: self->capacityP];

 return self;
}

//ppppp--set up the Scn properties and related variables (for ofpModelSwarm)--
-setScnId: (int) anId {
 self->scnId = anId;
 return self;
}

-(int) getScnId {
 return self->scnId;
}

-setTraceCode: (int) code {
 self->tC = code;
 return self;
}

-(int) getTC {
 return self->tC;
}

-setMfgCap: (char) capacity {
 self->withMfgCap = capacity;
 return self;
}

-(char) getMfgCap {
  return withMfgCap;
}

//--assign demand management policy
-assignDemandPolicy: (int) policy {
 self->demandPolicy = policy;
 return self;
}

-(int) getDemandPolicy {
 return self->demandPolicy;
}

//--assign coordination strategy to entity
-assignCoordIU: (int) strategy {
 self->coordIU = strategy;
 return self;
}
-assignCoordID: (int) strategy {
 self->coordID = strategy;
 return self;
}
-assignCoordSMC: (int) strategy {
 self->coordSMC = strategy;
 return self;
}
-assignCoordDBC: (int) strategy {
 self->coordDBC = strategy;
 return self;
}

//---assign the info share propagation layers
-assignUpPropLayer: (int) strategy {
 self->upPropLayer = strategy;
 return self;
}
-assignDownPropLayer: (int) strategy {
 self->downPropLayer = strategy;
 return self;
}

//-- get coordination strategy
-(int) getCoordIU {
 return self->coordIU;
}
-(int) getCoordID {
 return self->coordID;
}
-(int) getCoordSMC {
 return self->coordSMC;
}
-(int) getCoordDBC {
 return self->coordDBC;
}

//-- get info share propagation layers
-(int) getUpPropLayer {
 return self->upPropLayer;
}
-(int) getDownPropLayer {
 return self->downPropLayer;
}

//-- assign multisuppliers
-assignMultiSuppliers: (int) YesNo {
 self->multiSuppliers = YesNo;
 return self;
}

-(int) getMultiSuppliers {
 return multiSuppliers;
}

-assignDueDateTightness: (float) k {
 self->TWK = k;
return self;
}

-(float) getDueDateTightness {
 return self->TWK;
}

//--- set its agents --
-setOfpModel: (id) agent {
 self->ofpModel = agent;
 return self;
}

-addOneSupplier: (int) anId {
 IdNo * anE;
 anE = [IdNo create: [self getZone]];
 anE->no = anId;
 [self->suppliers addLast: anE];
 return self;
}


-addOneOrderCost: (float) aCost at: (int) offset {
 OrdCost * theOrdCost;
 IdNo * aSupplierId;

 theOrdCost = [OrdCost create: [self getZone]];
  if ([self->suppliers getCount] > 0) {
   aSupplierId = [self->suppliers atOffset: offset];
   theOrdCost->supplierId = aSupplierId->no;
   theOrdCost->cost = aCost;
   [self->orderCosts addLast: theOrdCost];
 }

 return self;
}

-addOneInvCost: (float) aCost types: (id) typeList at: (int) offset {
 InvCost * theInvCost;
 IdNo * aProdId;
 
 theInvCost = [InvCost create: [self getZone]];
 if ([typeList getCount] > 0) {
   aProdId = [typeList atOffset: offset];
   theInvCost->prodId = aProdId->no;
   theInvCost->unitCost = aCost;
   [self->inventoryCosts addLast: theInvCost];
 }
 return self;
}

-addOneCustomer: (int) anId {
 IdNo * anE;
 anE = [IdNo create: [self getZone]];
 anE->no = anId;
 [self->customers addLast: anE];
 return self;
}

-addOneOrderDelayT: (float) aTime {
 Ttime * aT;
 aT = [Ttime create: [self getZone]];
 aT->ttime = aTime;
 [self->orderDelayTs addLast: aT];
 return self;
}

-addOnePartDeliverT: (float) aTime {
 Ttime * aT;
 aT = [Ttime create: [self getZone]];
 aT->ttime = aTime;
 [self->partDeliverTs addLast: aT];
 return self;
}

-addOneProdId: (int) anId {
 IdNo * anE;
 anE = [IdNo create: [self getZone]];
 anE->no = anId;
 [self->prodIds addLast: anE];
 return self;
}

-addOneMfgProcT: (float) aTime {
 Ttime * aT;

 aT = [Ttime create: [self getZone]];
 aT->ttime = aTime;
 [self->mfgProcTs addLast: aT];
 return self;
}


-addOneProdDQ: (int) aDQ at: (int) offset{
 IdNo * aProdId;
 Part * aPart;
 InvCost * aCost;

 aPart = [Part create: [self getZone]];
 aPart->partQ = aDQ;
 if (self->withMfgCap == 'Y') {
   aProdId = [self->prodIds atOffset: offset];
   aPart->partId = aProdId->no;
 } 
 else {
   aCost = [self->inventoryCosts atOffset: offset];
   aPart->partId = aCost->prodId;
 }
 [self->inventoryDQs addLast: aPart];

 return self;
}

//sssss---- STATISTICS PROCESS (for ofpStatSwarm) --
-statProcess {
 float avgCycleT;
 float fulfilledRate;
 float avgInvCost;
 float capUtilization;
 float earlyD;
 float tardyD;
 float earlyW;
 float tardyW;

// cycle time
 avgCycleT = [self->invM getAvgCycleT]; 

// fulfilledRate
 fulfilledRate = [self->invM getFulfilledRate];

// inventory
 avgInvCost = [self->invM getAvgInvCost];
 
// capacity utilization
 if (self->withMfgCap == 'Y')
   capUtilization = [self->factory getCapUtilization];

// order due date performance
 tardyD = [self->invM getAvgDueDatePerformance: [self->invM getTardyDL]];
 earlyD = [self->invM getAvgDueDatePerformance: [self->invM getEarlyDL]];
 tardyW = [self->invM getAvgDueDatePerformance: [self->invM getTardyWL]];
 earlyW = [self->invM getAvgDueDatePerformance: [self->invM getEarlyWL]];
 
//
 if (self->withMfgCap != 'Y')
   capUtilization = -1.0;
 if (self->withMfgCap != 'S') {
   fprintf(freport, "%d  %4.2f %3.2f %6.2f %4.2f %4.2f %4.2f %4.2f %4.2f\n", self->scnId, avgCycleT, fulfilledRate, avgInvCost, capUtilization, earlyD, earlyW, tardyD, tardyW);
   // printf("%d %f %f %f %f %f %f %f %f\n", self->scnId, avgCycleT, fulfilledRate, avgInvCost, capUtilization, earlyD, earlyW, tardyD, tardyW);
  }
  return self;
}

//----get the capacity utilization statistics --
-(float) getCapUtilization {
 float capUtil, totalBusyT;
 
 totalBusyT= [self->factory getAccumuBusyTime];
 capUtil = totalBusyT / getCurrentTime();
 
 return capUtil;
}

//ddddd--------Methods for debugging --
-printSuppliers {
 IdNo * aS;
 id index;
 
 fprintf(fdebug, "SCNE = %d\n", [self getScnId]);
 index = [self->suppliers begin: [self getZone]];
 while ((aS = [index next])) { 
   fprintf(fdebug, "supplier = %d\n", aS->no);
 }
 [index drop];

 return self;
}

-printCustomers {
 IdNo * aC;
 id index;
 
 fprintf(fdebug, "SCNE = %d\n", [self getScnId]);
 index = [self->customers begin: [self getZone]];
 while ((aC = [index next])) { 
   fprintf(fdebug, "customer = %d\n", aC->no);
 }
 [index drop];

 return self;
}

//ffffff------Methods used for forecasting
-(int) belongToProdGroup: (int) prodId {
int belonging;
id index;
IdNo * aProd;
id <List> prodList;
  
  if ([self getMfgCap] == 'Y')
    prodList = [self getProdIds];
  else
    prodList = [self getInvTypes];

  index = [prodList begin: [self getZone]];
  belonging = FALSE;
  while ((aProd = [index next]) && (belonging == FALSE)) {
     if (aProd->no == prodId)
       belonging = TRUE;
  }

 return belonging;
}

//-----get a list of prodIds
-getProdGroup {
 id index;
 IdNo * aProd;
 id <List> prodList;
 id <List> aProdIds;

 if ([self getMfgCap] == 'Y')
   prodList = [self getProdIds];
 else
   prodList = [self getInvTypes];

 aProdIds = [List create: [self getZone]];
 index = [prodList begin: [self getZone]];
 while ((aProd = [index next]))
   [aProdIds addLast: aProd];
  
 [index drop];

 return aProdIds;
}

//rrrrrr--- Methods for reOrdering ---
//-----get order cost for an prodId for w/ and w/o mfgCap
-(float) getOrderCost: (int) prodId endItem: (int) isEndItem {
float theOrderCost;

 if (self->withMfgCap == 'N')
   theOrderCost = [self getOrderCostWithoutMfgCap: (int) prodId];
 else if (self->withMfgCap == 'Y') {
   if (isEndItem == Yes)   
     theOrderCost = [self getOrderCostWithMfgCap: (int) prodId];
   else if (isEndItem == No) 
     theOrderCost = [self getOrderCostWithoutMfgCap: (int) prodId];
 }
 return theOrderCost;
}

//--
-(float) getOrderCostWithoutMfgCap: (int) prodId {
id index, index1, index2;
OrdCost * ordCost;
int found = FALSE;
int aSupplierNo;
float theOrderCost = 0.0;
id <List> supplierList;
IdNo * aSupplierId;
int counter = 0;

 if (self->multiSuppliers == FALSE) {
   aSupplierNo = [self chooseSupplier: prodId]; // for single supplier case
   index = [self->orderCosts begin: [self getZone]];
   while ((ordCost = [index next]) && (found == FALSE)) {
     if (ordCost->supplierId == aSupplierNo) {
       theOrderCost = ordCost->cost;
       found = TRUE;
     }
   }
   [index drop];
 }
 else if (self->multiSuppliers == TRUE) {
   supplierList = [self chooseMultiSuppliers: prodId];
   index1 = [supplierList begin:[self getZone]];
   while ((aSupplierId = [index1 next])) {
     index2 = [self->orderCosts begin: [self getZone]];
     while ((ordCost = [index2 next]) && (found == FALSE)) {
       if (ordCost->supplierId == aSupplierId->no) {
         theOrderCost += ordCost->cost;
         counter++;
         found = TRUE;
       }
     }
     [index2 drop];
    }
    [index1 drop];
    theOrderCost /= counter;
  }

  return theOrderCost;
}

-(float) getOrderCostWithMfgCap: (int) prodId {
id <List> compL;
IdNo * aComp;
id index;
float ordCost = 0.0;

 compL = [ofpModel getCompList: prodId];
 index = [compL begin: [self getZone]];
 while ((aComp = [index next])) {
   ordCost += [self getOrderCostWithoutMfgCap: aComp->no];
 }
 [index drop];

 return ordCost;
}

//----get inventory cost for an partId
-(float) getInvCost: (int) prodId {
 id index;
 InvCost * aCost;
 int found = FALSE;
 float invC;

 index = [self->inventoryCosts begin: [self getZone]];
 while ((aCost = [index next]) && (found == FALSE)) {
   if (aCost->prodId == prodId) {
     invC = aCost->unitCost;
     found = TRUE;
   }
 }
 [index drop];

 return invC;     
}

//-- accept message to query actual demand 
-acceptComputeActualDemandQuery: (int)downProp from: (float) stime to: (float) etime prodSet: (id) prodList {
 id <List> actualDemandList;

 actualDemandList = [List create: [self getZone]];
 actualDemandList = [self->forecast computeActualDemand: downProp from: stime to: etime prodSet: prodList];
 
 return actualDemandList;
}

//---accept message, and give to ordM from other entity
-(float) acceptCommitDateQueryMessage: (Order *) anOrder level: (int) upProp {
 float commitDate;

 commitDate = [self->ordM  estimateCommitDate: anOrder level: upProp];
 return commitDate;
}

-(int) acceptDownDemandQuery: (DemandQuery *) aQuery {
 int amt = 0;

 if (self->withMfgCap == 'S')
   amt = 0;
 else if ([self belongToProdGroup: aQuery->prodId] == No)
   amt = 0;
 else
   amt = [self->reOrdM queryDemand: aQuery];

 return amt;
}

//--- find the endItem from component-
-(int) convertFromCompToEndItem: (int) aCompNo {
 int prodNo;
 id <List> endItemList;
 id <List> compList;
 id index1, index2;
 IdNo * aProd, *aCompId;
 int found = FALSE;

 endItemList = [self getProdIds];
 index1 = [endItemList begin: [self getZone]];
 while ((aProd = [index1 next]) && (found == FALSE)) {
   compList = [ofpModel  getCompList: aProd->no];
   index2 = [compList begin: [self getZone]];
   while ((aCompId = [index2 next]) && (found == FALSE)) {
      if (aCompId->no == aCompNo) {
        prodNo = aProd->no;
        found = TRUE;
      }
   } 
   [index2 drop];
 }
 [index1 drop];

 return prodNo;
}


//----obtain the inventoryDQs list --
-getInventoryDQs {
 id <List> invDQs;
 invDQs = [self->inventoryDQs copy: [self getZone]];
 return invDQs;
}

//------obtain the supplier List
-getSupplierList {
 id <List> supplierList;
 
 supplierList = [List create: [self getZone]];
 if ([self->suppliers getCount] > 0)
   supplierList = [self->suppliers copy: [self getZone]];
 
 return supplierList;
}

//------obtain the customer List
-getCustomerList {
 id <List> customerList;
 
 customerList = [List create: [self getZone]];
 if ([self->customers getCount] > 0)
   customerList = [self->customers copy: [self getZone]];
 
 return customerList;
}
//--- forward query, award, and bid to scnM --
-receiveQuery: (Query *) aQuery {
 [self->scnM receiveQuery: aQuery];
 return self;
}

-receiveAward: (Query *) anAward {
 [self->scnM receiveAward: anAward];
 return self;
}

-receiveBid: (Bid *) aBid {
 [self->scnM receiveBid: aBid];
 return self;
}

@end


