//Copyright (C) 1996 Fu-ren Lin
//ScnManagement.m

#import "ScnManagement.h"

@implementation ScnManagement
+create: aZone {
  ScnManagement * newScnM;
  newScnM = [super create: aZone];
  newScnM->bidList    = [List create: aZone]; 
  newScnM->queryList  = [List create: aZone]; 
  newScnM->queryRecList  = [List create: aZone]; 
  newScnM->queryInBuf = [List create: aZone];
  newScnM->bidInBuf   = [List create: aZone]; 
  newScnM->awardInBuf = [List create: aZone];
  newScnM->outSourcingList = [List create: aZone];
  newScnM->waitList   = [List create: aZone];
  return newScnM;
}

//----main process of scnM --
-process {
  [self generateQuery];
  [self processComingQuery];
  [self processReturningBid];
  [self processReturningAward];
  return self;
}

//- process the demand from inner agents stored in outSourcingList, and send
//   out to possible suppliers
-generateQuery {
Query * aQuery;
Order * anOrder;
 
 [traceAgent pHeader: [scnEntity getTC] value: "scnM=" fmt: Space];
 [traceAgent pInt: [scnEntity getTC] value: [scnEntity getScnId] fmt:NL];

  while ((anOrder = [self moveFromOutSourcingListToWaitList])) {
     aQuery = [Query create: [self getZone]];
     aQuery->queryId = self->qCnt;
     aQuery->prodId = anOrder->itemId;
     aQuery->prodQ  = anOrder->itemQ;
     if (anOrder->wantDate < 0)
       aQuery->prodNeedT = getCurrentTime();
     else
       aQuery->prodNeedT = anOrder->wantDate;
     aQuery->queryArrTime = getCurrentTime();
     aQuery->senderId = [scnEntity getScnId];
     [self->queryList addLast: aQuery];
     self->qCnt++;  //<------ assign the queryId
    
     [traceAgent pHeader: [scnEntity getTC] value: "broadcasting queries" fmt: Space];
     [traceAgent pQuery: [scnEntity getTC] value: aQuery];
  
     [self sendQueries: aQuery];
  }
  return self;
}

//- Receive the query from other entities, and issue the quote (bidding)
-processComingQuery {
 Query * inQuery;
 id index;

  if ([self->queryInBuf getCount] > 0) {
    index = [self->queryInBuf begin: [self getZone]];
    while ((inQuery = [index next])) {

      [traceAgent pHeader: [scnEntity getTC] value:"receive query" fmt: Space];
      [traceAgent pQuery: [scnEntity getTC] value: inQuery];
  
      if ([scnEntity belongToProdGroup: inQuery->prodId] == FALSE) 
        [self returnNullBid: inQuery];
      else
        [self reactToQuery: inQuery];  // **by sending a bid
      [index remove];  // remove the coming query
    }
    [index drop];
  }
  return self;
}

//- Receive the bids returned from other entities according to the
//   queries sent out previously
-processReturningBid {
Bid * returnBid;
Query * aLoser;
QueryRec * aQueryRec;
id index;

 if ([self->bidInBuf getCount] > 0) {
    index = [self->bidInBuf begin: [self getZone]];
    while ((returnBid = [index next])) {  

      [traceAgent pHeader: [scnEntity getTC] value:"receive bid" fmt: Space];
      [traceAgent pBid: [scnEntity getTC] value: returnBid];
  
      if ((aQueryRec = [self allocateQueryRec: returnBid])) {
        if ([self checkDoubleBid: aQueryRec from: returnBid] == FALSE)
          aQueryRec->returnQ++;
        [aQueryRec->returnBids addLast: returnBid];
        if (aQueryRec->returnQ == aQueryRec->sentQ) { 
            /* || ((getCurrentTime() -aQueryRec->querySentT) >= 
                                    self->considerationTime)) {*/
           [self evaluateAndRespondToBids: returnBid]; //**
           [self removeOneQueryRec: aQueryRec];
        }
      }
      else {
        aLoser = [self convertBidToQuery: returnBid];
        [self sendLoser: aLoser to: returnBid->senderId];
      }
      [index remove];
    }
    [index drop];
  }
  return self;
}

 //-- respond to award message (or a loser)
-processReturningAward { 
Query * returnAward;
id index;

 if ([self->awardInBuf getCount] > 0 ) {
   index = [self->awardInBuf begin: [self getZone]];
   while ((returnAward = [index next])) {
     [traceAgent pHeader: [scnEntity getTC] value:"receive award" fmt: Space];
     [traceAgent pAward: [scnEntity getTC] value: returnAward];
     
     if (returnAward->prodQ > 0)
       [self convertAwardToDemandOrder: returnAward];
     else
       [self removeBids: returnAward];

     [index remove];  //remove the coming award message
   }
   [index drop];
 }
 
 return self;
}

//--remove entries of bidlist according to loser message
-removeBids: (Query *) aLoser {
 id index;
 Bid * aBid;
  
 index = [self->bidList begin: [self getZone]];
 while ((aBid = [index next])) {
   if (aBid->queryId == aLoser->queryId)
     [index remove];
 }
 [index drop];

 return self;
}

//--remove one entry of queryRec --
-removeOneQueryRec: (QueryRec *) theQueryRec {
 id index;
 QueryRec * aQueryRec;
 int found = FALSE;

 index = [self->queryRecList begin: [self getZone]];
 while ((aQueryRec = [index next]) && (found == FALSE)) {
   if (aQueryRec->queryId == theQueryRec->queryId) {
     [index remove];
     found = TRUE;
   }
 }
 [index drop];

 return self;
}

//---convert the award to demandOrder for ordM --
-convertAwardToDemandOrder: (Query *) anAward {
 Order * demandOrder;
 
 if (anAward->prodQ > 0) {
   demandOrder = [self convertQueryToOrder: anAward];
   [ordAgent getOrder: demandOrder];
   
   [traceAgent pHeader: [scnEntity getTC] value:"award to OrdInBuf" fmt: Space];
   [traceAgent pOrder: [scnEntity getTC] value: demandOrder];
  
 }
 
 return self;
}

//----convert aBid to a query format 
-(Query *) convertBidToQuery: (Bid *) aBid {
 Query * aQuery;
 
 aQuery = [Query create: [self getZone]];
 aQuery->queryId = aBid->queryId;
 aQuery->prodId  = aBid->prodId;
 aQuery->prodQ   = aBid->prodQ;
 aQuery->queryArrTime = getCurrentTime()
                        + [scnEntity getDelayTime: aBid->senderId];
 aQuery->prodNeedT = aBid->prodDeliverT;
 aQuery->senderId = [scnEntity getScnId];

 return aQuery;
}

//--- take a demandOrder to consider in allocating resources
-acceptDemandOrder: (Order *) demandOrder {
 id index;
 Order * anOrder;
 int found = FALSE;

 index = [self->outSourcingList begin: [self getZone]];
 while ((anOrder = [index next]) && (found == FALSE)) {
   if (anOrder->dueDate > demandOrder->dueDate) {
     [index addBefore: demandOrder];
     found = TRUE;
   }
 }
 [index drop];

 if (found == FALSE) {
   [self->outSourcingList addLast: demandOrder];
 }

  [traceAgent pHeader: [scnEntity getTC] value:"accept an Order" fmt: Space];
  [traceAgent pOrder: [scnEntity getTC] value: demandOrder];
  
 return self;
}


//---move top of outSourcingList to the end of waitList --
-(Order *) moveFromOutSourcingListToWaitList {
 id index;
 Order * anOrder, *waitOrder;

 index = [self->outSourcingList begin:[self getZone]];
 if ((anOrder = [index next])) {
   waitOrder = [anOrder copy: [self getZone]];
   [self->waitList addLast: waitOrder];
   [index remove];
 }
 [index drop];
 return anOrder;
}

//----send queries to possible entities --
-sendQueries:  (Query *) aQuery {
 IdNo * aSupplierId;
 SCNEntitySwarm * aSupplierE;
 id <List> supplierList;
 id index;
 QueryRec * queryRec;
 Query * outQuery;

 supplierList = [scnEntity getSupplierList];
 if ([supplierList getCount] > 0) {
   queryRec = [QueryRec create: [self getZone]];
   queryRec->queryId = aQuery->queryId;
   queryRec->querySentT = getCurrentTime();
   queryRec->sentQ   = 0;
   queryRec->returnQ = 0;
   queryRec->returnBids = [List create:[self getZone]];
   queryRec->winnerList = [List create:[self getZone]];
 }
 else 
   return self;

 index = [supplierList begin: [self getZone]];
 while ((aSupplierId = [index next])) {
   aSupplierE = [ofpModel allocateEntity: aSupplierId->no];
   aQuery->queryArrTime += [scnEntity getDelayTime: aSupplierId->no];
   outQuery = [Query create: [self getZone]];
   outQuery->queryId = aQuery->queryId;
   outQuery->prodId = aQuery->prodId;
   outQuery->prodQ  = aQuery->prodQ;
   outQuery->queryArrTime = aQuery->queryArrTime;
   outQuery->prodNeedT = aQuery->prodNeedT;
   outQuery->senderId = aQuery->senderId;

   [aSupplierE receiveQuery: outQuery];
   queryRec->sentQ++;
 }
 [index drop];
 [self->queryRecList addLast: queryRec];
 return self;
}

//----receive award from other entity and store in awardInbuf list
-receiveQuery: (Query *) inQuery {
 Query * aQuery;
 id index;
 int found = FALSE;

 index = [self->queryInBuf begin: [self getZone]];
 while ((aQuery = [index next]) && (found == FALSE)) {
   if (aQuery->queryArrTime > inQuery->queryArrTime) {
     [index addBefore: inQuery];
     found = TRUE;
   }
 }
 [index drop];

 if (found == FALSE) {
   [self->queryInBuf addLast: inQuery];
 }
 
 return self;
}
//----receive an Award from other entity and store in AwardInbuf list
-receiveAward: (Query *) inAward {
 Query * anAward;
 id index;
 int found = FALSE;

 index = [self->awardInBuf begin: [self getZone]];
 while ((anAward = [index next]) && (found == FALSE)) {
   if (anAward->queryArrTime > inAward->queryArrTime) {
     [index addBefore: inAward];
     found = TRUE;
   }
 }
 [index drop];

 if (found == FALSE) {
   [self->awardInBuf addLast: inAward];
 }
 
 return self;
}

//----receive a bid from other entity and store in bidInbuf list
-receiveBid: (Bid *) inBid {
 Bid * aBid;
 id index;
 int found = FALSE;

 index = [self->bidInBuf begin: [self getZone]];
 while ((aBid = [index next]) && (found == FALSE)) {
   if (aBid->bidArrTime > inBid->bidArrTime) {
     [index addBefore: inBid];
     found = TRUE;
   }
 }
 [index drop];

 if (found == FALSE) {
   [self->bidInBuf addLast: inBid];
 }
 
 return self;
}

//-- send out an invalid bid because of no capacity or not in prod group
-returnNullBid: (Query *) aQuery {
 Bid * outBid;
 SCNEntitySwarm * requesterE;

 outBid = [Bid create: [self getZone]];
 outBid->queryId = aQuery->queryId;
 outBid->prodId  = aQuery->prodId;
 outBid->prodQ   = 0;
 outBid->marginalC  = Infinite;
 outBid->prodDeliverT = Infinite;
 outBid->bidArrTime = getCurrentTime() + [scnEntity getDelayTime: aQuery->senderId to: [scnEntity getScnId]];
 outBid->senderId = [scnEntity getScnId];
 requesterE = [SCNEntitySwarm create: [self getZone]];
 requesterE = [ofpModel allocateEntity: aQuery->senderId];
 [requesterE receiveBid: outBid];

 return self;
}

//----react to query from other entity -- calcurate "marginal costs",
// response query as a returning bid
-reactToQuery: (Query *) aQuery {
 Order * anOrder, * theOrder;
 int needToOrder;
 Bid * outBid, * aBid;
 int aReceiverNo;
 SCNEntitySwarm * requesterE;
 float deliverTime;
 int accumuOrder, currentInv, inOrdering, avail;

 // prepare outgoing Bid 
 aBid = [Bid create: [self getZone]];
 aBid->queryId = aQuery->queryId;
 aBid->prodId  = aQuery->prodId;
 aBid->bidArrTime = getCurrentTime() + [scnEntity getDelayTime
                      :aQuery->senderId to: [scnEntity getScnId]];
 aBid->senderId = [scnEntity getScnId];
 aBid->prodQ = aQuery->prodQ;

 aBid->prodDeliverT = 0.0;
 aBid->marginalC    = 0.0;

 //specify receiver Id and deliverTime ....
 aReceiverNo = aQuery->senderId;
 deliverTime = [scnEntity getDeliverTime: aReceiverNo];
 requesterE = [ofpModel allocateEntity:  aReceiverNo];

 // calcurate marginal cost and availTime ...
 anOrder = [self convertQueryToOrder: aQuery];
 needToOrder = [ordAgent checkAvailInv: anOrder];
 if (needToOrder == 0) {   // from inv or receiving
    aBid->marginalC = 0.0; //<---------------* unique bid
    anOrder->dueDate = [ordAgent estimateCommitDate: anOrder level: [scnEntity getUpPropLayer]];
    aBid->prodDeliverT = anOrder->dueDate + deliverTime;
    outBid = [aBid copy: [self getZone]];
    [requesterE receiveBid: outBid];
    [self->bidList addLast: outBid];
 }
 else {                    // from inv, receiving, and re-ordering
   accumuOrder = [invAgent sumOfDemandList: anOrder];
   currentInv = [invAgent checkInv: anOrder->itemId];
   inOrdering = [ordAgent checkInOrdering: anOrder];
   avail = anOrder->itemQ - needToOrder;
   theOrder = [anOrder copy: [self getZone]];
   theOrder->itemQ = avail;
   
   if (avail > 0) {         //<---------------* 1st bid
     theOrder = [anOrder copy: [self getZone]];
     theOrder->itemQ = avail;
     aBid->marginalC = 0.0;
     aBid->prodQ = avail;
     anOrder->dueDate = [ordAgent estimateCommitDate: theOrder level: [scnEntity getUpPropLayer]];
     aBid->prodDeliverT = anOrder->dueDate + deliverTime;
     outBid = [aBid copy: [self getZone]];
     [requesterE receiveBid: outBid];
     [self->bidList addLast: outBid];
   }   
                             //<----------------* 2nd bid
   aBid->marginalC = [self calcurateMarginalCost: aQuery with: needToOrder];
   aBid->prodQ = needToOrder;
   anOrder->dueDate = [ordAgent estimateCommitDate: anOrder level: [scnEntity getUpPropLayer]];
   aBid->prodDeliverT = anOrder->dueDate + deliverTime;
   outBid = [aBid copy: [self getZone]];
   [requesterE receiveBid: outBid];
   [self->bidList addLast: outBid];
 }
 
 return self;
}

-(Order *) convertQueryToOrder: (Query *) aQuery {
 Order * anOrder;

 anOrder = [Order create: [self getZone]];
 anOrder->itemId = aQuery->prodId;
 anOrder->itemQ = aQuery->prodQ;
 anOrder->ordArrTime = aQuery->queryArrTime;
 anOrder->wantDate = aQuery->prodNeedT;
 anOrder->dueDate = aQuery->prodNeedT;
 anOrder->senderId = aQuery->senderId;

return anOrder;
}

-(float) calcurateMarginalCost: (Query *) aQuery with: (int) needQ {
float marginalCost;
  
  marginalCost = [scnEntity getOrderCost: aQuery->prodId endItem: Yes];
  return marginalCost;
}

//--- allocate the queryRec entry for the return bids
-(QueryRec *) allocateQueryRec: (Bid *) returnBid {
id index;
QueryRec * sentQuery;
int found = FALSE;
int offset = -1;

  index = [self->queryRecList begin: [self getZone]];
  while ((sentQuery = [index next]) && (found == FALSE)) {
    offset++; 
    if (sentQuery->queryId == returnBid->queryId)
      found = TRUE;
  }
  [index drop];
  if (found == TRUE)
   sentQuery = [self->queryRecList atOffset: offset];

  return sentQuery;
}

//-----for splitting a query into different returning quotes
-(int) checkDoubleBid: (QueryRec *) aQueryRec from: (Bid *) returnBid {
int duplicated = FALSE;
id index;
Bid * aBid;

 index = [aQueryRec->returnBids begin: [self getZone]];
 while ((aBid = [index next]) && (duplicated == FALSE)) {
    if (aBid->senderId == returnBid->senderId)
       duplicated = TRUE;
 }
 [index drop];
 return duplicated;
}
 
//----evaluate all returned bids and choose the best one.
-evaluateAndRespondToBids: (Bid *) returnBid {
 id index;
 QueryRec * aQueryRec, * theQueryRec;
 Query * aQuery, * theQuery;
 int found;
 id <List> evaList;
 Bid * aBid;
 Evaluation * anEval;
 int Qleft;
 float OIratio = OIRATIO;
 float invCost, opCost;
 float min;
 int offset, point;
 int complete;
 int currentBidder;
 IdNo * aWinner;

  //allocate the queryRec
  found = FALSE;
  index = [self->queryRecList begin: [self getZone]];
  while ((aQueryRec = [index next]) && (found == FALSE)) {
    if (aQueryRec->queryId == returnBid->queryId) {
      theQueryRec = [index remove];
      found = TRUE;
    }
  }
  [index drop];
  
  // allocate the original query
  found = FALSE;
  index = [self->queryList begin: [self getZone]];
  while ((aQuery = [index next]) && (found == FALSE)) {
    if (aQuery->queryId == returnBid->queryId) {
      theQuery = [index remove];
      found = TRUE;
    }
  }
  [index drop];
  
 // evaluate the cost and time gap, then write to evaList
  evaList = [List create: [self getZone]];
  index = [theQueryRec->returnBids begin: [self getZone]];
  while ((aBid = [index next])) {
    if (aBid->prodQ > 0) {
      anEval = [Evaluation create: [self getZone]];
      anEval->prodQ   = aBid->prodQ;
      anEval->mCost   = aBid->marginalC;
      anEval->timeGap = aBid->prodDeliverT - theQuery->prodNeedT;
      invCost = [scnEntity getInvCost: aBid->prodId];
      if (anEval->timeGap > 0)  { //delay
        opCost = invCost * OIratio;
        anEval->utility = anEval->mCost + (anEval->timeGap * opCost);
      }
      else
        anEval->utility = (anEval->timeGap * invCost) - anEval->mCost;

      anEval->supplierId = aBid->senderId;
      [evaList addLast: anEval];
    }
  }
  [index drop];
 
 [traceAgent pHeader: [scnEntity getTC] value: "evaluate bid from query" fmt: Space];
 [traceAgent pQuery: [scnEntity getTC] value: theQuery];
  
 // choose the supplier
 Qleft = theQuery->prodQ;
 min = Infinite;
 complete = FALSE;
 while ((Qleft > 0) && (complete == FALSE)) {
   offset = -1;
   point = -1;
   index = [evaList begin: [self getZone]];
   while ((anEval = [index next])) {

   [traceAgent pHeader: [scnEntity getTC] value: "evaluation" fmt: Space];
   [traceAgent pEval: [scnEntity getTC] value: anEval];

     offset++;
     if (anEval->utility < min) {
       min = anEval->utility;
       point = offset;
     }
   }
   [index drop];

   if (point >= 0) {
     anEval = [evaList atOffset: point];
     Qleft = [self sendAward: theQuery to: anEval max: Qleft];
     anEval->utility = Infinite;

     currentBidder = anEval->supplierId;
     index = [theQueryRec->winnerList begin: [self getZone]];
     found = FALSE;
     while ((aWinner = [index next]) && (found == FALSE)) {
       if (aWinner->no == currentBidder) {
         found = TRUE;
       }
     }
     [index drop];
     if (found == FALSE) {
       aWinner = [IdNo create: [self getZone]];
       aWinner->no = currentBidder;
       [theQueryRec->winnerList addLast: aWinner];
     }
   }

   if ((Qleft == 0) || (point < 0)) 
     complete = TRUE;
 }  // end of while (( Qleft > 0) ...
 
 [self sendLosersToAllBidders: theQueryRec->returnBids]; // to stop the bidding

 if (Qleft > 0) { // some products are not awarded
   [self reSendQueries: theQuery with: Qleft];
 }
 
 return self;
}

-reSendQueries: (Query *) aQuery with: (int) prodQ {
 aQuery->prodQ = prodQ;
 [self sendQueries: aQuery];
 return self;
} 
 
//--- send out award to bidders
-(int) sendAward: (Query *) aQuery to: (Evaluation *) anEval max: (int) maxQ {
int stillNeed;
SCNEntitySwarm * aSupplierE;
Query * outAward;

  if (maxQ >= anEval->prodQ) {
    stillNeed = maxQ - anEval->prodQ;
    aQuery->prodQ = anEval->prodQ;
  }
  else {
    stillNeed = 0;
    aQuery->prodQ = maxQ;
  }
  
  aSupplierE = [ofpModel allocateEntity: anEval->supplierId];
  aQuery->queryArrTime += [scnEntity getDelayTime: anEval->supplierId];
  outAward = [Query create: [self getZone]];
  outAward->queryId = aQuery->queryId;
  outAward->prodId = aQuery->prodId;
  outAward->prodQ  = aQuery->prodQ;
  outAward->queryArrTime = aQuery->queryArrTime;
  outAward->prodNeedT = aQuery->prodNeedT;
  outAward->senderId = aQuery->senderId;
  
  [aSupplierE receiveAward: outAward];    // send the award to a supplier
/*
  demandOrder = [Order create: [self getZone]];  // add into the demandList
  deamndOrder = [self convertQueryToOrder: outAward];
  [invAgent addToDemandList: demandOrder];
*/
 return stillNeed;
}

//---send losers to all of bidders ---
-sendLosersToAllBidders: (id) returnBidList {
 id index;
 Bid * aBid;
 Query * aLoser;
 
 index = [returnBidList begin: [self getZone]];
 while ((aBid = [index next])) {
   aLoser = [self convertBidToQuery: aBid];
   [self sendLoser: aLoser to: aBid->senderId];
 }
 
 return self;
}

// check to see if it is a winner
-(int) isAWinner: (int) sender by: (id) winnerList {
 int aWinner;
 id index;
 IdNo * aWin;
 int found = FALSE;

 index = [winnerList begin: [self getZone]];
 while ((aWin = [index next]) && (found == FALSE)) {
   if (aWin->no == sender)
     found = TRUE;
 }
 [index drop];
 if (found == TRUE)
   aWinner = Yes;
 else
   aWinner = No;

 return aWinner;
}

//---send loser informaton --
-sendLoser: (Query *) aLoser to: (int) aBidderNo {
 SCNEntitySwarm * aSupplierE;

 aSupplierE = [ofpModel allocateEntity: aBidderNo];
 aLoser->prodQ = 0;
 [aSupplierE receiveAward: aLoser];

 return self;
}


//---- initialize the parameters --
-initBiddingCounter {
 self->qCnt = 1;
 self->considerationTime = BiddingTime;
 return self;
}


//-----set up agents --
-setInvAgent: (id) agent {
 invAgent = agent;
 return self;
}

-setOrdAgent: (id) agent {
 ordAgent = agent;
 return self;
}

-setMrpAgent: (id) agent {
 mrpAgent = agent;
 return self;
}

-setPrdPlanningAgent: (id) agent {
 prdPlanningAgent = agent;
 return self;
}

-setScnEntity: (id) agent {
 scnEntity = agent;
 return self;
}

-setOfpModel: (id) agent {
 ofpModel = agent;
 return self;
}


-setTraceAgent: (id) agent {
 traceAgent = agent;
 return self;
}

@end
